from __future__ import annotations

import functools
import inspect
import logging
import os
import string
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar, overload

from githarbor.core.models import (
    Branch,
    Commit,
    Issue,
    Label,
    PullRequest,
    Release,
    Tag,
    User,
    Workflow,
    WorkflowRun,
)
from githarbor.exceptions import ResourceNotFoundError


if TYPE_CHECKING:
    from collections.abc import Callable

    from github.AuthenticatedUser import AuthenticatedUser
    from github.NamedUser import NamedUser


T = TypeVar("T")
P = ParamSpec("P")
TOKEN = os.getenv("GITHUB_TOKEN")
logger = logging.getLogger(__name__)


def handle_github_errors(
    error_msg_template: str,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator to handle GitHub API exceptions consistently.

    Args:
        error_msg_template: Message template with format placeholders

    Example:
        @handle_github_errors("Could not fetch branch {branch_name}")
        def get_branch(self, branch_name: str) -> Branch:
            ...
    """
    # Extract field names from the template string
    parser = string.Formatter()
    param_names = {
        field_name
        for _, field_name, _, _ in parser.parse(error_msg_template)
        if field_name and field_name != "error"
    }

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            from github.GithubException import GithubException

            # Extract parameter values from args/kwargs based on function signature
            sig = inspect.signature(func)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()
            params = {
                name: bound_args.arguments[name]
                for name in param_names
                if name in bound_args.arguments
            }

            try:
                return func(*args, **kwargs)
            except GithubException as e:
                msg = error_msg_template.format(**params, error=str(e))
                raise ResourceNotFoundError(msg) from e

        return wrapper

    return decorator


def download_from_github(
    org: str,
    repo: str,
    path: str | os.PathLike[str],
    destination: str | os.PathLike[str],
    username: str | None = None,
    token: str | None = None,
    recursive: bool = False,
):
    import fsspec
    import upath

    token = token or TOKEN
    if token and not username:
        token = None
    dest = upath.UPath(destination)
    dest.mkdir(exist_ok=True, parents=True)
    fs = fsspec.filesystem("github", org=org, repo=repo)
    logger.info("Copying files from Github: %s", path)
    files = fs.ls(str(path))
    fs.get(files, dest.as_posix(), recursive=recursive)


@overload
def create_user_model(gh_user: None) -> None: ...


@overload
def create_user_model(gh_user: NamedUser | AuthenticatedUser) -> User: ...


def create_user_model(gh_user: NamedUser | AuthenticatedUser | None) -> User | None:
    """Create User model from GitHub user object."""
    from github.NamedUser import NamedUser

    if not gh_user:
        return None
    return User(
        username=gh_user.login,
        name=gh_user.name,
        email=gh_user.email,
        avatar_url=gh_user.avatar_url,
        created_at=gh_user.created_at,
        bio=gh_user.bio,
        location=gh_user.location,
        company=gh_user.company,
        url=gh_user.html_url,
        followers=gh_user.followers,
        following=gh_user.following,
        public_repos=gh_user.public_repos,
        blog=gh_user.blog,
        twitter_username=gh_user.twitter_username
        if isinstance(gh_user, NamedUser)
        else None,
        hireable=gh_user.hireable,
        gravatar_id=gh_user.gravatar_id,
    )


def create_label_model(gh_label: Any) -> Label:
    """Create Label model from GitHub label object."""
    return Label(
        name=gh_label.name,
        color=gh_label.color,
        description=gh_label.description or "",
        url=gh_label.url,
    )


def create_pull_request_model(pr: Any) -> PullRequest:
    return PullRequest(
        number=pr.number,
        title=pr.title,
        description=pr.body or "",
        state=pr.state,
        source_branch=pr.head.ref,
        target_branch=pr.base.ref,
        created_at=pr.created_at,
        updated_at=pr.updated_at,
        merged_at=pr.merged_at,
        closed_at=pr.closed_at,
        author=create_user_model(pr.user),
        assignees=[create_user_model(a) for a in pr.assignees if a],
        labels=[create_label_model(lbl) for lbl in pr.labels],
        merged_by=create_user_model(pr.merged_by),
        review_comments_count=pr.review_comments,
        commits_count=pr.commits,
        additions=pr.additions,
        deletions=pr.deletions,
        changed_files=pr.changed_files,
        mergeable=pr.mergeable,
        url=pr.html_url,
    )


def create_issue_model(issue: Any) -> Issue:
    return Issue(
        number=issue.number,
        title=issue.title,
        description=issue.body or "",
        state=issue.state,
        created_at=issue.created_at,
        updated_at=issue.updated_at,
        closed_at=issue.closed_at,
        closed=issue.state == "closed",
        author=create_user_model(issue.user),
        assignee=create_user_model(issue.assignee),
        labels=[create_label_model(lbl) for lbl in issue.labels],
        comments_count=issue.comments,
        url=issue.html_url,
        milestone=issue.milestone.title if issue.milestone else None,
    )


def create_commit_model(commit: Any) -> Commit:
    return Commit(
        sha=commit.sha,
        message=commit.commit.message,
        created_at=commit.commit.author.date,
        author=create_user_model(commit.author)
        or User(
            username="",
            name=commit.commit.author.name,
            email=commit.commit.author.email,
        ),
        committer=create_user_model(commit.committer),
        url=commit.html_url,
        stats={
            "additions": commit.stats.additions,
            "deletions": commit.stats.deletions,
            "total": commit.stats.total,
        },
        parents=[p.sha for p in commit.parents],
        # verified=commit.commit.verification.verified,
        files_changed=[f.filename for f in commit.files],
    )


def create_release_model(release: Any) -> Release:
    return Release(
        tag_name=release.tag_name,
        name=release.title,
        description=release.body or "",
        created_at=release.created_at,
        published_at=release.published_at,
        draft=release.draft,
        prerelease=release.prerelease,
        author=User(
            username=release.author.login,
            name=release.author.name,
            avatar_url=release.author.avatar_url,
        )
        if release.author
        else None,
        assets=[
            {
                "name": asset.name,
                "url": asset.browser_download_url,
                "size": asset.size,
                "download_count": asset.download_count,
                "created_at": asset.created_at,
                "updated_at": asset.updated_at,
            }
            for asset in release.assets
        ],
        url=release.html_url,
        target_commitish=release.target_commitish,
    )


def create_workflow_model(workflow: Any) -> Workflow:
    """Create Workflow model from GitHub workflow object."""
    # raw_prefix = f"https://raw.githubusercontent.com/{self._owner}/{self._name}/"
    return Workflow(
        id=str(workflow.id),
        name=workflow.name,
        path=workflow.path,
        state=workflow.state,
        created_at=workflow.created_at,
        updated_at=workflow.updated_at,
        description=workflow.name,  # GitHub API doesn't provide separate description
        triggers=[],  # Would need to parse the workflow file to get triggers
        disabled=workflow.state.lower() == "disabled",
        last_run_at=None,  # Not directly available from the API
        badge_url=workflow.badge_url,
        # definition=f"{raw_prefix}{self.default_branch}/{workflow.path}",
    )


def create_workflow_run_model(run: Any) -> WorkflowRun:
    """Create WorkflowRun model from GitHub workflow run object."""
    return WorkflowRun(
        id=str(run.id),
        name=run.name or run.display_title,
        workflow_id=str(run.workflow_id),
        status=run.status,
        conclusion=run.conclusion,
        branch=run.head_branch,
        commit_sha=run.head_sha,
        url=run.html_url,
        created_at=run.created_at,
        updated_at=run.updated_at,
        started_at=run.run_started_at,
        completed_at=run.run_attempt_started_at,
        run_number=run.run_number,
        jobs_count=len(list(run.jobs())),
        logs_url=run.logs_url,
    )


def create_tag_model(tag: Any) -> Tag:
    """Create Tag model from GitHub tag object."""
    return Tag(
        name=tag.name,
        sha=tag.commit.sha if hasattr(tag, "commit") else tag.object.sha,
        message=tag.message if hasattr(tag, "message") else None,
        created_at=tag.tagger.date if hasattr(tag, "tagger") else None,
        author=create_user_model(tag.tagger) if hasattr(tag, "tagger") else None,
        url=tag.url if hasattr(tag, "url") else None,
    )


def create_branch_model(branch: Any) -> Branch:
    """Create Branch model from GitHub branch object."""
    last_commit = branch.commit
    return Branch(
        name=branch.name,
        sha=branch.commit.sha,
        protected=branch.protected,
        default=False,  # This needs to be set by the caller
        protection_rules=(
            {
                "required_reviews": branch.get_required_status_checks(),
                "dismiss_stale_reviews": branch.get_required_pull_request_reviews(),
                "require_code_owner_reviews": branch.get_required_signatures(),
            }
            if branch.protected
            else None
        ),
        last_commit_date=last_commit.commit.author.date,
        last_commit_message=last_commit.commit.message,
        last_commit_author=create_user_model(last_commit.author),
    )


def create_file_model(content: Any) -> dict[str, Any]:
    """Create file info dictionary from GitHub content object."""
    return {
        "name": content.name,
        "path": content.path,
        "sha": content.sha,
        "size": content.size,
        "type": content.type,
        "url": content.html_url,
        "download_url": content.download_url,
        "encoding": content.encoding if hasattr(content, "encoding") else None,
    }



from typing import Any, ClassVar

import pytest

from githarbor.core import base
from githarbor.exceptions import RepositoryNotFoundError
from githarbor.registry import RepoRegistry


class DummyRepository(base.BaseRepository):
    url_patterns: ClassVar[list[str]] = ["test.com"]

    def __init__(self, owner, repo, **kwargs: Any):
        self.kwargs = kwargs

    @classmethod
    def from_url(cls, url: str, **kwargs: Any) -> base.BaseRepository:
        return cls("test", "test")


def test_register_repo():
    # Clear existing registrations
    RepoRegistry._repo_classes = {}  # Changed from _repos to _repos

    @RepoRegistry.register("test")
    class DummyRepositoryRegistered(DummyRepository):
        pass

    assert "test" in RepoRegistry._repo_classes
    assert RepoRegistry._repo_classes["test"] == DummyRepositoryRegistered


def test_create_repo():
    # Register test repo
    RepoRegistry._repo_classes = {"test": DummyRepository}

    # Create by name
    repo = RepoRegistry.create("test", owner="test", repo="test", token="test-token")
    assert isinstance(repo._repository, DummyRepository)
    assert repo._repository.kwargs["token"] == "test-token"

    # Invalid repo
    with pytest.raises(RepositoryNotFoundError):
        RepoRegistry.create("invalid")


def test_from_url():
    # Register test repo
    RepoRegistry._repo_classes = {"test": DummyRepository}

    # Create from URL
    repo = RepoRegistry.from_url("https://test.com/owner/repo", token="test-token")
    assert isinstance(repo._repository, DummyRepository)

    # Invalid URL
    with pytest.raises(RepositoryNotFoundError):
        RepoRegistry.from_url("https://invalid.com/owner/repo")


def test_get_repo_class_for_url():
    # Register test repo
    RepoRegistry._repo_classes = {"test": DummyRepository}

    # Get repo class
    repo_class = RepoRegistry.get_repo_class_for_url("https://test.com/owner/repo")
    assert repo_class == DummyRepository

    # Invalid URL
    repo_class = RepoRegistry.get_repo_class_for_url("https://invalid.com/owner/repo")
    assert repo_class is None


def test_get_registered_repo_classes():
    # Register test repos
    RepoRegistry._repo_classes = {"test1": DummyRepository, "test2": DummyRepository}

    repos = RepoRegistry.get_registered_repo_classes()
    assert set(repos) == {"test1", "test2"}



"""Base model for Githarbor."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from datetime import datetime


@dataclass
class User:
    """GitHub user model."""

    username: str
    """Username of the user."""
    name: str | None = None
    """Full name of the user."""
    email: str | None = None
    """Email address of the user."""
    avatar_url: str | None = None
    """URL of the user's avatar image."""
    created_at: datetime | None = None
    """Date and time when the user account was created."""
    bio: str | None = None
    """Biography/description of the user."""
    location: str | None = None
    """Geographic location of the user."""
    company: str | None = None
    """Company/organization the user belongs to."""
    state: str | None = None
    """State of the user (active / inactive)."""
    locked: bool | None = None
    """Whether user is locked."""
    followers: int | None = None
    """Number of followers."""
    following: int | None = None
    """Number of users being followed."""
    public_repos: int | None = None
    """Number of public repositories."""
    url: str | None = None
    """URL to user's profile."""
    blog: str | None = None
    """Personal blog of this user"""
    twitter_username: str | None = None
    """X / Twitter username of this user"""
    hireable: bool | None = None
    """Whether user is available for hiring"""
    gravatar_id: str | None = None
    """Gravatar id for this user."""
    is_admin: bool | None = None
    """Whether the user has admin privileges."""
    last_activity_on: datetime | None = None
    """Date of last activity."""
    linkedin: str | None = None
    """LinkedIn username."""
    skype: str | None = None
    """Skype username."""


@dataclass
class Label:
    name: str
    """Name of the label."""
    color: str = ""  # Hex color code
    """Hex color code for the label."""
    description: str = ""
    """Description of what the label represents."""
    created_at: datetime | None = None
    """Date and time when the label was created."""
    updated_at: datetime | None = None
    """Date and time when the label was last updated."""
    url: str | None = None
    """URL to the label."""


@dataclass
class Comment:
    id: str
    """Unique identifier for the comment."""
    body: str
    """Content of the comment."""
    author: User
    """User who wrote the comment."""
    created_at: datetime
    """Date and time when the comment was created."""
    updated_at: datetime | None = None
    """Date and time when the comment was last updated."""
    reactions: dict[str, int] = field(default_factory=dict)
    """Dictionary of reaction types and their counts."""
    reply_to: str | None = None  # ID of parent comment if this is a reply
    """ID of parent comment if this is a reply."""
    url: str | None = None
    """URL to the comment."""


@dataclass
class Branch:
    name: str
    """Name of the branch."""
    sha: str
    """SHA hash of the latest commit on this branch."""
    protected: bool = False
    """Whether the branch has protection rules enabled."""
    default: bool = False
    """Whether this is the default branch of the repository."""
    created_at: datetime | None = None
    """Date and time when the branch was created."""
    updated_at: datetime | None = None
    """Date and time when the branch was last updated."""
    last_commit_date: datetime | None = None
    """Date of the last commit."""
    last_commit_message: str | None = None
    """Message of the last commit."""
    last_commit_author: User | None = None
    """Author of the last commit."""
    protection_rules: dict[str, Any] | None = None
    """Branch protection settings if any."""


@dataclass
class PullRequest:
    """Pull request model."""

    number: int
    """Pull request number."""
    title: str
    """Title of the pull request."""
    source_branch: str
    """Branch containing the changes."""
    target_branch: str
    """Branch to merge into."""
    description: str = ""
    """Description/body of the pull request."""
    state: str | None = None
    """Current state (open/closed/merged)."""
    created_at: datetime | None = None
    """Creation timestamp."""
    updated_at: datetime | None = None
    """Last update timestamp."""
    merged_at: datetime | None = None
    """Merge timestamp if merged."""
    closed_at: datetime | None = None
    """Closing timestamp if closed."""
    author: User | None = None
    """PR creator."""
    assignees: list[User] = field(default_factory=list)
    """Assigned reviewers."""
    labels: list[Label] = field(default_factory=list)
    """Applied labels."""
    comments: list[Comment] = field(default_factory=list)
    """PR comments."""
    merged_by: User | None = None
    """User who merged the PR."""
    review_comments_count: int | None = None
    """Number of review comments."""
    commits_count: int | None = None
    """Number of commits."""
    additions: int | None = None
    """Lines added."""
    deletions: int | None = None
    """Lines deleted."""
    changed_files: int | None = None
    """Number of files changed."""
    mergeable: bool | None = None
    """Whether PR can be merged."""
    url: str | None = None
    """URL to the PR."""


@dataclass
class Issue:
    number: int
    """Unique identifier for the issue."""
    title: str
    """Title of the issue."""
    description: str = ""
    """Description of the issue."""
    state: str = "open"
    """State of the issue."""
    author: User | None = None
    """User who created the issue."""
    assignee: User | None = None
    """User who is assigned to the issue."""
    labels: list[Label] = field(default_factory=list)
    """List of labels attached to the issue."""
    created_at: datetime | None = None
    """Date and time when the issue was created."""
    updated_at: datetime | None = None
    """Date and time when the issue was last updated."""
    closed_at: datetime | None = None
    """Date and time when the issue was closed."""
    closed: bool = False
    """Indicates if the issue is closed."""
    comments_count: int = 0
    """Number of comments."""
    url: str | None = None
    """URL to the issue."""
    milestone: str | None = None
    """Associated milestone."""


@dataclass
class Commit:
    sha: str
    """Unique identifier for the commit."""
    message: str
    """Commit message."""
    author: User | None = None
    """User who authored the commit."""
    created_at: datetime | None = None
    """Date and time when the commit was authored."""
    committer: User | None = None
    """User who committed the changes."""
    url: str | None = None
    """URL to the commit details."""
    stats: dict[str, int] = field(default_factory=dict)
    """Commit statistics."""
    parents: list[str] = field(default_factory=list)
    """List of parent commit SHAs."""
    verified: bool = False
    """Signature verification status."""
    files_changed: list[str] = field(default_factory=list)
    """Changed file paths."""


@dataclass
class Tag:
    """Model representing a repository tag."""

    name: str
    """Name of the tag."""
    sha: str
    """SHA hash of the commit this tag points to."""
    message: str | None = None
    """Tag message if provided."""
    created_at: datetime | None = None
    """Date and time when the tag was created."""
    author: User | None = None
    """User who created the tag."""
    url: str | None = None
    """URL to view the tag."""
    verified: bool = False
    """Whether the tag signature is verified."""


@dataclass
class Workflow:
    id: str
    """Unique identifier for the workflow."""
    name: str
    """Name of the workflow."""
    path: str
    """Path to the workflow file in the repository."""
    state: str
    """State of the workflow."""
    created_at: datetime | None = None
    """Date and time when the workflow was created."""
    updated_at: datetime | None = None
    """Date and time when the workflow was last updated."""
    description: str = ""
    """Description of the workflow."""
    triggers: list[str] = field(default_factory=list)
    """List of triggers that can start the workflow."""
    disabled: bool = False
    """Indicates if the workflow is disabled."""
    last_run_at: datetime | None = None
    """Date and time when the workflow was last run."""
    badge_url: str | None = None
    """URL to the badge image for the workflow."""
    definition: str | None = None
    """Content of the workflow definition file."""
    runs_count: int = 0
    """Total number of runs."""
    success_count: int = 0
    """Successful runs count."""


@dataclass
class WorkflowRun:
    id: str
    """Unique identifier for the workflow run."""
    name: str
    """Name of the workflow run."""
    workflow_id: str
    """ID of the parent workflow."""
    status: str
    """Current status of the workflow run."""
    conclusion: str | None
    """Final conclusion of the workflow run."""
    branch: str | None = None
    """Branch the workflow was run on."""
    commit_sha: str | None = None
    """SHA of the commit that triggered the workflow."""
    url: str | None = None
    """URL to view the workflow run."""
    created_at: datetime | None = None
    """Date and time when the workflow run was created."""
    updated_at: datetime | None = None
    """Date and time when the workflow run was last updated."""
    started_at: datetime | None = None
    """Date and time when the workflow run started."""
    completed_at: datetime | None = None
    """Date and time when the workflow run completed."""
    run_number: int | None = None
    """Sequential run number."""
    jobs_count: int | None = None
    """Total jobs count."""
    logs_url: str | None = None
    """Logs URL."""


@dataclass
class Release:
    """Model representing a repository release."""

    tag_name: str
    """Tag name for the release."""
    name: str
    """Name/title of the release."""
    description: str = ""
    """Description/body of the release."""
    created_at: datetime | None = None
    """Date and time when the release was created."""
    published_at: datetime | None = None
    """Date and time when the release was published."""
    draft: bool = False
    """Whether this is a draft release."""
    prerelease: bool = False
    """Whether this is a pre-release."""
    author: User | None = None
    """User who created the release."""
    assets: list[dict[str, Any]] = field(default_factory=list)
    """List of assets attached to the release."""
    url: str | None = None
    """URL to view the release."""
    target_commitish: str | None = None
    """The branch/tag/commit the release targets."""
    download_count: int = 0
    """Total downloads."""



"""Azure DevOps helper functions and decorators."""

from __future__ import annotations

import functools
import inspect
import logging
import os
import string
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar, overload

from githarbor.core.models import (
    Commit,
    Issue,
    Label,
    PullRequest,
    Release,
    Tag,
    User,
    Workflow,
    WorkflowRun,
)
from githarbor.exceptions import ResourceNotFoundError


if TYPE_CHECKING:
    from collections.abc import Callable

    from azure.devops.v7_1.git.models import GitCommit, GitPullRequest
    from azure.devops.v7_1.work_item_tracking.models import WorkItem as AzureWorkItem


logger = logging.getLogger(__name__)
T = TypeVar("T")
P = ParamSpec("P")
TOKEN = os.getenv("AZURE_DEVOPS_PAT")


def handle_azure_errors(
    error_msg_template: str,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator to handle Azure DevOps API exceptions consistently.

    Args:
        error_msg_template: Message template with format placeholders

    Example:
        @handle_azure_errors("Could not fetch branch {branch_name}")
        def get_branch(self, branch_name: str) -> Branch:
            ...
    """
    parser = string.Formatter()
    param_names = {
        field_name
        for _, field_name, _, _ in parser.parse(error_msg_template)
        if field_name and field_name != "error"
    }

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            from azure.devops.exceptions import AzureDevOpsServiceError

            sig = inspect.signature(func)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()
            params = {
                name: bound_args.arguments[name]
                for name in param_names
                if name in bound_args.arguments
            }

            try:
                return func(*args, **kwargs)
            except AzureDevOpsServiceError as e:
                msg = error_msg_template.format(**params, error=str(e))
                raise ResourceNotFoundError(msg) from e

        return wrapper

    return decorator


def download_from_azure(
    organization: str,
    project: str,
    repo: str,
    path: str | os.PathLike[str],
    destination: str | os.PathLike[str],
    token: str | None = None,
    recursive: bool = False,
):
    """Download files from Azure DevOps repository.

    Args:
        organization: Azure DevOps organization name
        project: Project name
        repo: Repository name
        path: Path to file/folder to download
        destination: Local destination path
        token: Personal access token
        recursive: Whether to download recursively
    """
    from azure.devops.connection import Connection
    from msrest.authentication import BasicAuthentication
    import upath

    token = token or TOKEN
    if not token:
        msg = "Azure DevOps PAT token is required"
        raise ValueError(msg)

    dest = upath.UPath(destination)
    dest.mkdir(exist_ok=True, parents=True)

    credentials = BasicAuthentication("", token)
    connection = Connection(
        base_url=f"https://dev.azure.com/{organization}",
        creds=credentials,
    )
    git_client = connection.clients.get_git_client()

    logger.info("Downloading files from Azure DevOps: %s", path)
    content = git_client.get_item_content(
        repository_id=repo,
        path=str(path),
        project=project,
        download=True,
    )

    if recursive:
        msg = "Recursive download not yet implemented for Azure DevOps"
        raise NotImplementedError(msg)

    file_dest = dest / upath.UPath(path).name
    file_dest.write_bytes(content)


@overload
def create_user_model(azure_user: None) -> None: ...


@overload
def create_user_model(azure_user: Any) -> User: ...


def create_user_model(azure_user: Any) -> User | None:
    """Create User model from Azure DevOps identity object."""
    if not azure_user:
        return None
    return User(
        username=azure_user.get("uniqueName", ""),
        name=azure_user.get("displayName", ""),
        email=azure_user.get("mailAddress", ""),
        avatar_url=azure_user.get("imageUrl", ""),
    )


def create_label_model(azure_label: Any) -> Label:
    """Create Label model from Azure DevOps label object."""
    return Label(
        name=azure_label.name,
        color=azure_label.color or "000000",
        description=azure_label.description or "",
        url="",  # Azure DevOps doesn't provide direct URLs for labels
    )


def create_pull_request_model(pr: GitPullRequest) -> PullRequest:
    """Create PullRequest model from Azure DevOps pull request object."""
    return PullRequest(
        number=pr.pull_request_id or 0,
        title=pr.title or "unknown",
        description=pr.description or "",
        state=pr.status,
        source_branch=pr.source_ref_name.split("/")[-1],  # type: ignore
        target_branch=pr.target_ref_name.split("/")[-1],  # type: ignore
        created_at=pr.creation_date,
        updated_at=None,  # Not directly provided by Azure DevOps
        merged_at=pr.closed_date if pr.status == "completed" else None,
        closed_at=pr.closed_date if pr.status in ["abandoned", "completed"] else None,
        author=create_user_model(pr.created_by),
        assignees=[create_user_model(r) for r in pr.reviewers or []],
        labels=[],  # Would need additional processing
        merged_by=create_user_model(pr.closed_by) if pr.closed_by else None,
        url=pr.url,
    )


def create_issue_model(work_item: AzureWorkItem) -> Issue:
    """Create Issue model from Azure DevOps work item object."""
    fields = work_item.fields
    return Issue(
        number=work_item.id,
        title=fields["System.Title"],
        description=fields.get("System.Description", ""),
        state=fields["System.State"],
        created_at=fields["System.CreatedDate"],
        updated_at=fields.get("System.ChangedDate"),
        closed_at=None,  # Not directly provided
        closed=fields["System.State"] in ["Closed", "Resolved"],
        author=create_user_model(fields.get("System.CreatedBy")),
        assignee=create_user_model(fields.get("System.AssignedTo")),
        labels=[],  # Would need additional processing
        url=work_item.url,
    )


def create_commit_model(commit: GitCommit) -> Commit:
    """Create Commit model from Azure DevOps commit object."""
    return Commit(
        sha=commit.commit_id,
        message=commit.comment,
        created_at=commit.author.date,
        author=User(
            username=commit.author.name,
            name=commit.author.name,
            email=commit.author.email,
        ),
        committer=User(
            username=commit.committer.name,
            name=commit.committer.name,
            email=commit.committer.email,
        ),
        url=commit.url,
        stats={
            "additions": commit.change_counts.add if commit.change_counts else 0,
            "deletions": commit.change_counts.delete if commit.change_counts else 0,
            "total": commit.change_counts.edit if commit.change_counts else 0,
        },
        parents=[p.commit_id for p in commit.parents] if commit.parents else [],
    )


def create_release_model(tag: Any) -> Release:
    """Create Release model from Azure DevOps tag object."""
    return Release(
        tag_name=tag.name,
        name=tag.name,
        description=tag.message or "",
        created_at=tag.commit.committer.date,
        published_at=tag.commit.committer.date,
        draft=False,  # Azure doesn't have draft concept for tags
        prerelease=False,  # Azure doesn't have prerelease concept for tags
        author=User(
            username=tag.commit.committer.name,
            name=tag.commit.committer.name,
            email=tag.commit.committer.email,
        ),
        url=None,  # Azure doesn't provide direct URLs for tags
        target_commitish=tag.commit.commit_id,
        assets=[],  # Azure tags don't have assets
    )


def create_workflow_model(definition: Any) -> Workflow:
    """Create Workflow model from Azure DevOps build definition object."""
    return Workflow(
        id=str(definition.id),
        name=definition.name,
        path=definition.path or "",
        state="enabled" if definition.quality == "enabled" else "disabled",
        created_at=None,  # Not provided by Azure DevOps
        updated_at=None,  # Not provided by Azure DevOps
        description=definition.description or "",
        triggers=[t.type for t in definition.triggers] if definition.triggers else [],
        disabled=definition.quality != "enabled",
        badge_url=None,  # Not directly provided
    )


def create_workflow_run_model(build: Any) -> WorkflowRun:
    """Create WorkflowRun model from Azure DevOps build object."""
    return WorkflowRun(
        id=str(build.id),
        name=build.definition.name,
        workflow_id=str(build.definition.id),
        status=build.status,
        conclusion=build.result or "",
        branch=build.source_branch.split("/")[-1],
        commit_sha=build.source_version,
        url=build.url,
        created_at=build.queue_time,
        updated_at=None,  # Not directly provided
        started_at=build.start_time,
        completed_at=build.finish_time,
        run_number=build.build_number,
        jobs_count=None,  # Would need additional API calls
        logs_url=build.logs.url if build.logs else None,
    )


def create_tag_model(tag: Any) -> Tag:
    return Tag(
        name=tag.name,
        sha=tag.commit.commit_id,
        message=tag.message or "",
        created_at=tag.commit.committer.date,
        author=User(
            username=tag.commit.committer.name,
            name=tag.commit.committer.name,
            email=tag.commit.committer.email,
        ),
        url=None,  # Azure DevOps doesn't provide direct URLs for tags
    )



"""Module containing the repository registry."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar
from weakref import WeakValueDictionary

from githarbor.core.proxy import Repository
from githarbor.core.user_proxy import Owner
from githarbor.exceptions import RepositoryNotFoundError


if TYPE_CHECKING:
    from githarbor.core.base import BaseOwner, BaseRepository


class RepoRegistry:
    """Registry for repository implementations."""

    _repo_classes: ClassVar[dict[str, type[BaseRepository]]] = {}
    _owner_classes: ClassVar[dict[str, type[BaseOwner]]] = {}
    _repo_instances: ClassVar[WeakValueDictionary[str, Repository]] = (
        WeakValueDictionary()
    )
    _owner_instances: ClassVar[WeakValueDictionary[str, Owner]] = WeakValueDictionary()

    @classmethod
    def register(cls, name: str):
        """Decorator to register a repository class.

        Args:
            name: Name to register the repository under.
        """

        def decorator(repo_class: type[BaseRepository]) -> type[BaseRepository]:
            cls._repo_classes[name] = repo_class
            return repo_class

        return decorator

    @classmethod
    def create(cls, name: str, **kwargs: Any) -> Repository:
        """Create a proxy-wrapped repository instance by name.

        Args:
            name: Name of the repository type.
            **kwargs: Repository-specific configuration.

        Returns:
            Proxy-wrapped repository instance.

        Raises:
            RepositoryNotFoundError: If repository type is not found or creation fails.
        """
        if not (repo_class := cls._repo_classes.get(name)):
            msg = f"Repository type {name} not found"
            raise RepositoryNotFoundError(msg)
        return Repository(repo_class(**kwargs))

    @classmethod
    def get(cls, url: str, **kwargs: Any) -> Repository:
        """Get cached repository instance or create new one."""
        # Generate cache key from URL and relevant kwargs
        cache_key = f"{url}:{hash(frozenset(kwargs.items()))}"

        if cache_key in cls._repo_instances:
            return cls._repo_instances[cache_key]

        repo = cls.from_url(url, **kwargs)
        cls._repo_instances[cache_key] = repo
        return repo

    @classmethod
    def from_url(cls, url: str, **kwargs: Any) -> Repository:
        """Create a proxy-wrapped repository instance from a URL.

        Args:
            url: Repository URL.
            **kwargs: Repository-specific configuration.

        Returns:
            Proxy-wrapped repository instance.

        Raises:
            RepositoryNotFoundError: If no implementation supports the URL.
        """
        url = url.removesuffix(".git")
        for repo_class in cls._repo_classes.values():
            if repo_class.supports_url(url):
                return Repository(repo_class.from_url(url, **kwargs))

        msg = f"No repository implementation found for URL: {url}"
        raise RepositoryNotFoundError(msg)

    @classmethod
    def get_repo_class_for_url(cls, url: str) -> type[BaseRepository] | None:
        """Get the repository class that can handle the given URL.

        Args:
            url: Repository URL.

        Returns:
            Repository class or None if no implementation supports the URL.
        """
        return next(
            (
                repo_class
                for repo_class in cls._repo_classes.values()
                if repo_class.supports_url(url)
            ),
            None,
        )

    @classmethod
    def get_registered_repo_classes(cls) -> list[str]:
        """Get a list of all registered repository types.

        Returns:
            List of registered repository type names.
        """
        return list(cls._repo_classes)

    @classmethod
    def clear_cache(cls) -> None:
        """Clear the instance cache."""
        cls._repo_instances.clear()

    @classmethod
    def register_owner(cls, name: str):
        """Decorator to register an owner class."""

        def decorator(owner_class: type[BaseOwner]) -> type[BaseOwner]:
            cls._owner_classes[name] = owner_class
            return owner_class

        return decorator

    @classmethod
    def get_owner(cls, url: str, **kwargs: Any) -> Owner:
        """Get cached owner instance or create new one."""
        cache_key = f"{url}:{hash(frozenset(kwargs.items()))}"

        if cache_key in cls._owner_instances:
            return cls._owner_instances[cache_key]

        owner = cls.owner_from_url(url, **kwargs)
        cls._owner_instances[cache_key] = owner
        return owner

    @classmethod
    def owner_from_url(cls, url: str, **kwargs: Any) -> Owner:
        """Create an owner instance from a URL."""
        url = url.removesuffix(".git")
        for owner_class in cls._owner_classes.values():
            if owner_class.supports_url(url):
                base_owner = owner_class.from_url(url, **kwargs)
                return Owner(base_owner)

        msg = f"No owner implementation found for URL: {url}"
        raise RepositoryNotFoundError(msg)


registry = RepoRegistry()



from __future__ import annotations

import asyncio
import functools
from typing import TYPE_CHECKING, Any, Literal, ParamSpec, TypeVar

from githarbor.registry import RepoRegistry


if TYPE_CHECKING:
    from collections.abc import Callable
    from datetime import datetime
    import os

    from githarbor.core.base import BaseRepository, IssueState, PullRequestState
    from githarbor.core.models import (
        Branch,
        Commit,
        Issue,
        PullRequest,
        Release,
        Tag,
        User,
        Workflow,
        WorkflowRun,
    )

P = ParamSpec("P")
T = TypeVar("T")


def make_sync(async_func: Callable[P, T]) -> Callable[P, T]:
    """Convert an async function to sync using asyncio.run()."""

    @functools.wraps(async_func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return asyncio.run(async_func(*args, **kwargs))  # type: ignore[arg-type]

    return wrapper


async def get_repo_user_async(url: str) -> User:
    """Get repository owner information.

    Args:
        url: Repository URL
    """
    repo = RepoRegistry.get(url)
    return await repo.get_repo_user_async()


async def get_branch_async(url: str, name: str) -> Branch:
    """Get information about a specific repository branch.

    Args:
        url: Repository URL
        name: Branch name
    """
    repo = RepoRegistry.get(url)
    return await repo.get_branch_async(name)


async def get_pull_request_async(url: str, number: int) -> PullRequest:
    """Get information about a specific pull request.

    Args:
        url: Repository URL
        number: Pull request number
    """
    repo = RepoRegistry.get(url)
    return await repo.get_pull_request_async(number)


async def list_pull_requests_async(
    url: str, *, state: PullRequestState = "open"
) -> list[PullRequest]:
    """List repository pull requests.

    Args:
        url: Repository URL
        state: Pull request state filter ('open', 'closed', 'all')
    """
    repo = RepoRegistry.get(url)
    return await repo.list_pull_requests_async(state)


async def get_issue_async(url: str, issue_id: int) -> Issue:
    """Get information about a specific issue.

    Args:
        url: Repository URL
        issue_id: Issue number
    """
    repo = RepoRegistry.get(url)
    return await repo.get_issue_async(issue_id)


async def list_issues_async(url: str, *, state: IssueState = "open") -> list[Issue]:
    """List repository issues.

    Args:
        url: Repository URL
        state: Issue state filter ('open', 'closed', 'all')
    """
    repo = RepoRegistry.get(url)
    return await repo.list_issues_async(state)


async def get_commit_async(url: str, sha: str) -> Commit:
    """Get information about a specific commit.

    Args:
        url: Repository URL
        sha: Commit SHA
    """
    repo = RepoRegistry.get(url)
    return await repo.get_commit_async(sha)


async def list_commits_async(
    url: str,
    *,
    branch: str | None = None,
    since: datetime | None = None,
    until: datetime | None = None,
    author: str | None = None,
    path: str | None = None,
    max_results: int | None = None,
) -> list[Commit]:
    """List repository commits with optional filters.

    Args:
        url: Repository URL
        branch: Filter by branch name
        since: Only show commits after this date
        until: Only show commits before this date
        author: Filter by author
        path: Filter by file path
        max_results: Maximum number of results
    """
    repo = RepoRegistry.get(url)
    return await repo.list_commits_async(
        branch=branch,
        since=since,
        until=until,
        author=author,
        path=path,
        max_results=max_results,
    )


async def get_workflow_async(url: str, workflow_id: str) -> Workflow:
    """Get information about a specific workflow.

    Args:
        url: Repository URL
        workflow_id: Workflow identifier
    """
    repo = RepoRegistry.get(url)
    return await repo.get_workflow_async(workflow_id)


async def list_workflows_async(url: str) -> list[Workflow]:
    """List repository workflows.

    Args:
        url: Repository URL
    """
    repo = RepoRegistry.get(url)
    return await repo.list_workflows_async()


async def get_workflow_run_async(url: str, run_id: str) -> WorkflowRun:
    """Get information about a specific workflow run.

    Args:
        url: Repository URL
        run_id: Workflow run identifier
    """
    repo = RepoRegistry.get(url)
    return await repo.get_workflow_run_async(run_id)


async def download_async(
    url: str,
    path: str | os.PathLike[str],
    destination: str | os.PathLike[str],
    *,
    recursive: bool = False,
) -> None:
    """Download repository content.

    Args:
        url: Repository URL
        path: Path to download
        destination: Where to save the downloaded content
        recursive: Whether to download recursively
    """
    repo = RepoRegistry.get(url)
    await repo.download_async(path, destination, recursive)


async def search_commits_async(
    url: str,
    query: str,
    *,
    branch: str | None = None,
    path: str | None = None,
    max_results: int | None = None,
) -> list[Commit]:
    """Search repository commits.

    Args:
        url: Repository URL
        query: Search query string
        branch: Filter by branch name
        path: Filter by file path
        max_results: Maximum number of results
    """
    repo = RepoRegistry.get(url)
    return await repo.search_commits_async(query, branch, path, max_results)


async def get_contributors_async(
    url: str,
    *,
    sort_by: Literal["commits", "name", "date"] = "commits",
    limit: int | None = None,
) -> list[User]:
    """Get repository contributors.

    Args:
        url: Repository URL
        sort_by: How to sort the contributors
        limit: Maximum number of contributors to return
    """
    repo = RepoRegistry.get(url)
    return await repo.get_contributors_async(sort_by, limit)


async def get_languages_async(url: str) -> dict[str, int]:
    """Get repository language statistics.

    Args:
        url: Repository URL
    """
    repo = RepoRegistry.get(url)
    return await repo.get_languages_async()


async def compare_branches_async(
    url: str,
    base: str,
    head: str,
    *,
    include_commits: bool = True,
    include_files: bool = True,
    include_stats: bool = True,
) -> dict[str, Any]:
    """Compare two branches.

    Args:
        url: Repository URL
        base: Base branch name
        head: Head branch name
        include_commits: Whether to include commit information
        include_files: Whether to include changed files
        include_stats: Whether to include statistics
    """
    repo = RepoRegistry.get(url)
    return await repo.compare_branches_async(
        base, head, include_commits, include_files, include_stats
    )


async def get_latest_release_async(
    url: str,
    *,
    include_drafts: bool = False,
    include_prereleases: bool = False,
) -> Release:
    """Get latest repository release.

    Args:
        url: Repository URL
        include_drafts: Whether to include draft releases
        include_prereleases: Whether to include pre-releases
    """
    repo = RepoRegistry.get(url)
    return await repo.get_latest_release_async(include_drafts, include_prereleases)


async def list_releases_async(
    url: str,
    *,
    include_drafts: bool = False,
    include_prereleases: bool = False,
    limit: int | None = None,
) -> list[Release]:
    """List repository releases.

    Args:
        url: Repository URL
        include_drafts: Whether to include draft releases
        include_prereleases: Whether to include pre-releases
        limit: Maximum number of releases to return
    """
    repo = RepoRegistry.get(url)
    return await repo.list_releases_async(include_drafts, include_prereleases, limit)


async def get_release_async(url: str, tag: str) -> Release:
    """Get release by tag.

    Args:
        url: Repository URL
        tag: Release tag name
    """
    repo = RepoRegistry.get(url)
    return await repo.get_release_async(tag)


async def get_tag_async(url: str, name: str) -> Tag:
    """Get tag information.

    Args:
        url: Repository URL
        name: Tag name
    """
    repo = RepoRegistry.get(url)
    return await repo.get_tag_async(name)


async def list_tags_async(url: str) -> list[Tag]:
    """List repository tags.

    Args:
        url: Repository URL
    """
    repo = RepoRegistry.get(url)
    return await repo.list_tags_async()


async def list_repositories_async(url: str) -> list[BaseRepository]:
    """List repositories owned by user/organization.

    Args:
        url: Owner URL
    """
    owner = RepoRegistry.get_owner(url)
    return await owner.list_repositories_async()


async def create_repository_async(
    url: str,
    name: str,
    description: str = "",
    private: bool = False,
) -> BaseRepository:
    """Create a new repository.

    Args:
        url: Owner URL
        name: Repository name
        description: Repository description
        private: Whether to create a private repository
    """
    owner = RepoRegistry.get_owner(url)
    return await owner.create_repository_async(name, description, private)


async def get_user_async(url: str) -> User:
    """Get user information.

    Args:
        url: User URL
    """
    owner = RepoRegistry.get_owner(url)
    return await owner.get_user_async()


async def delete_repository_async(url: str, name: str) -> None:
    """Delete a repository.

    Args:
        url: Owner URL
        name: Repository name to delete
    """
    owner = RepoRegistry.get_owner(url)
    await owner.delete_repository_async(name)


async def create_pull_request_async(
    url: str,
    title: str,
    body: str,
    head_branch: str,
    base_branch: str,
    draft: bool = False,
) -> PullRequest:
    """Create a new pull request.

    Args:
        url: Repository URL
        title: Pull request title
        body: Pull request description
        head_branch: Source branch containing the changes
        base_branch: Target branch for the changes
        draft: Whether to create a draft pull request

    Returns:
        Newly created pull request
    """
    repo = RepoRegistry.get(url)
    return await repo.create_pull_request_async(
        title=title,
        body=body,
        head_branch=head_branch,
        base_branch=base_branch,
        draft=draft,
    )


async def create_pull_request_from_diff_async(
    url: str,
    title: str,
    body: str,
    base_branch: str,
    diff: str,
    head_branch: str | None = None,
    draft: bool = False,
) -> PullRequest:
    """Create a pull request from a diff string.

    Args:
        url: Repository URL
        title: Pull request title
        body: Pull request description
        base_branch: Target branch for the changes
        diff: Git diff string
        head_branch: Name of the branch to create. Auto-generated if not provided.
        draft: Whether to create a draft pull request

    Returns:
        Created pull request
    """
    repo = RepoRegistry.get(url)
    return await repo.create_pull_request_from_diff_async(
        title=title,
        body=body,
        base_branch=base_branch,
        diff=diff,
        head_branch=head_branch,
        draft=draft,
    )


async def create_branch_async(
    url: str,
    name: str,
    base_commit: str,
) -> Branch:
    """Create a new branch at the specified commit.

    Args:
        url: Repository URL
        name: Name of the branch to create
        base_commit: SHA of the commit to base the branch on

    Returns:
        Created branch
    """
    repo = RepoRegistry.get(url)
    return await repo.create_branch_async(
        name=name,
        base_commit=base_commit,
    )


get_repo_user = make_sync(get_repo_user_async)
get_branch = make_sync(get_branch_async)
create_branch = make_sync(create_branch_async)
get_pull_request = make_sync(get_pull_request_async)
list_pull_requests = make_sync(list_pull_requests_async)
get_issue = make_sync(get_issue_async)
list_issues = make_sync(list_issues_async)
get_commit = make_sync(get_commit_async)
list_commits = make_sync(list_commits_async)
get_workflow = make_sync(get_workflow_async)
list_workflows = make_sync(list_workflows_async)
get_workflow_run = make_sync(get_workflow_run_async)
download = make_sync(download_async)
search_commits = make_sync(search_commits_async)
get_contributors = make_sync(get_contributors_async)
get_languages = make_sync(get_languages_async)
compare_branches = make_sync(compare_branches_async)
get_latest_release = make_sync(get_latest_release_async)
list_releases = make_sync(list_releases_async)
get_release = make_sync(get_release_async)
get_tag = make_sync(get_tag_async)
list_tags = make_sync(list_tags_async)
create_pull_request = make_sync(create_pull_request_async)
create_pull_request_from_diff = make_sync(create_pull_request_async)
list_repositories = make_sync(list_repositories_async)
create_repository = make_sync(create_repository_async)
get_user = make_sync(get_user_async)
delete_repository = make_sync(delete_repository_async)


def setup_env(env: Any) -> None:
    """Used as extension point for the jinjarope environment.

    Args:
        env: The jinjarope environment to extend
    """
    methods = [
        # Async methods
        get_repo_user_async,
        get_branch_async,
        create_branch_async,
        get_pull_request_async,
        list_pull_requests_async,
        get_issue_async,
        list_issues_async,
        get_commit_async,
        list_commits_async,
        get_workflow_async,
        list_workflows_async,
        get_workflow_run_async,
        download_async,
        search_commits_async,
        get_contributors_async,
        get_languages_async,
        compare_branches_async,
        get_latest_release_async,
        list_releases_async,
        get_release_async,
        get_tag_async,
        list_tags_async,
        create_pull_request_async,
        create_pull_request_from_diff_async,
        list_repositories_async,
        create_repository_async,
        get_user_async,
        delete_repository_async,
        # Sync methods
        get_repo_user,
        get_branch,
        create_branch,
        get_pull_request,
        list_pull_requests,
        get_issue,
        list_issues,
        get_commit,
        list_commits,
        get_workflow,
        list_workflows,
        get_workflow_run,
        download,
        search_commits,
        get_contributors,
        get_languages,
        compare_branches,
        get_latest_release,
        list_releases,
        get_release,
        get_tag,
        list_tags,
        create_pull_request,
        create_pull_request_from_diff,
        list_repositories,
        create_repository,
        get_user,
        delete_repository,
    ]

    funcs = {func.__name__: func for func in methods}
    # Special case for download which has a different name in the interface
    funcs["download_from_repo"] = download
    funcs["download_from_repo_async"] = download_async

    # Register as both globals and filters
    env.globals |= funcs
    env.filters |= funcs


if __name__ == "__main__":

    async def main():
        workflows = await list_workflows_async("https://github.com/phil65/githarbor")
        print(workflows)

    import asyncio

    asyncio.run(main())



from __future__ import annotations

import importlib.util
from typing import TYPE_CHECKING, Any

from githarbor.exceptions import RepositoryNotFoundError
from githarbor.registry import RepoRegistry


if TYPE_CHECKING:
    from githarbor.core.proxy import Repository


if importlib.util.find_spec("github"):
    from githarbor.providers.github_provider.repository import GitHubRepository

    RepoRegistry.register("github")(GitHubRepository)

if importlib.util.find_spec("aiogithubapi"):
    from githarbor.providers.aiogithubapi_provider.repository import AioGitHubRepository

    RepoRegistry.register("aiogithubapi")(AioGitHubRepository)

if importlib.util.find_spec("gitlab"):
    from githarbor.providers.gitlab_provider.repository import GitLabRepository

    RepoRegistry.register("gitlab")(GitLabRepository)

if importlib.util.find_spec("giteapy"):
    from githarbor.providers.gitea_provider.repository import GiteaRepository

    RepoRegistry.register("gitea")(GiteaRepository)

if importlib.util.find_spec("azure"):
    from githarbor.providers.azure_provider.repository import AzureRepository

    RepoRegistry.register("azure")(AzureRepository)

if importlib.util.find_spec("githubkit"):
    from githarbor.providers.githubkit_provider.repository import GitHubKitRepository

    RepoRegistry.register("githubkit")(GitHubKitRepository)

# if importlib.util.find_spec("atlassian"):
#     from githarbor.providers.bitbucket_provider.repository import BitbucketRepository

#     RepoRegistry.register("bitbucket")(BitbucketRepository)


if importlib.util.find_spec("git"):
    # registered last so that it's the fallback, since we allow upaths this would
    # also pick up all other URLs.
    from githarbor.providers.local_provider.repository import LocalRepository

    RepoRegistry.register("local")(LocalRepository)


def create_repository(url: str, **kwargs: Any) -> Repository:
    """Create a proxy-wrapped repository instance from a URL.

    Args:
        url: The repository URL (e.g. 'https://github.com/owner/repo')
        **kwargs: Repository-specific configuration (tokens, credentials, etc.)

    Returns:
        Repository: Proxy-wrapped repository instance

    Raises:
        RepositoryNotFoundError: If the URL isn't supported or no repository found

    Example:
        >>> repo = create_repository('https://github.com/owner/repo', token='my-token')
        >>> issues = repo.list_issues()
    """
    try:
        return RepoRegistry.from_url(url, **kwargs)
    except Exception as e:
        msg = f"Failed to create repository from {url}: {e!s}"
        raise RepositoryNotFoundError(msg) from e


if __name__ == "__main__":
    repo = create_repository(".")
    print(repo._repository)



from __future__ import annotations

from datetime import datetime

import pytest

from githarbor.core.models import (
    Branch,
    Comment,
    Commit,
    Issue,
    Label,
    PullRequest,
    User,
    Workflow,
    WorkflowRun,
)


@pytest.fixture
def mock_user():
    return User(
        username="test-user",
        name="Test User",
        email="test@example.com",
        avatar_url="https://example.com/avatar.png",
        created_at=datetime(2023, 1, 1),
        bio="Test bio",
        location="Test Location",
        company="Test Company",
    )


@pytest.fixture
def mock_label():
    return Label(
        name="bug",
        color="#ff0000",
        description="Bug label",
        created_at=datetime(2023, 1, 1),
        updated_at=datetime(2023, 1, 1),
    )


@pytest.fixture
def mock_comment(mock_user):
    return Comment(
        id="1",
        body="Test comment",
        author=mock_user,
        created_at=datetime(2023, 1, 1),
        updated_at=datetime(2023, 1, 1),
        reactions={"thumbs_up": 1, "heart": 2},
        reply_to=None,
    )


@pytest.fixture
def mock_commit(mock_user):
    return Commit(
        sha="abc123def456",
        message="Test commit message",
        author=mock_user,
        created_at=datetime(2023, 1, 1),
        committer=mock_user,
        url="https://github.com/owner/repo/commit/abc123def456",
        stats={"additions": 10, "deletions": 5},
        parents=["xyz789"],
    )


@pytest.fixture
def mock_issue(mock_user, mock_label):
    return Issue(
        title="Test Issue",
        number=42,
        description="Test issue description",
        state="open",
        author=mock_user,
        assignee=mock_user,
        labels=[mock_label],
        created_at=datetime(2023, 1, 1),
        updated_at=datetime(2023, 1, 1),
        closed_at=None,
        closed=False,
    )


@pytest.fixture
def mock_workflow():
    return Workflow(
        id="workflow-1",
        name="CI Pipeline",
        path=".github/workflows/ci.yml",
        state="active",
        created_at=datetime(2023, 1, 1),
        updated_at=datetime(2023, 1, 1),
        description="Continuous Integration workflow",
        triggers=["push", "pull_request"],
        disabled=False,
        last_run_at=datetime(2023, 1, 1),
    )


@pytest.fixture
def mock_workflow_run():
    return WorkflowRun(
        id="run-1",
        name="CI Pipeline Run",
        workflow_id="workflow-1",
        status="completed",
        conclusion="success",
        branch="main",
        commit_sha="abc123def456",
        url="https://github.com/owner/repo/actions/runs/1",
        created_at=datetime(2023, 1, 1),
        updated_at=datetime(2023, 1, 1),
        started_at=datetime(2023, 1, 1),
        completed_at=datetime(2023, 1, 1),
    )


@pytest.fixture
def mock_branch() -> Branch:
    return Branch(
        name="feature-branch",
        sha="abc123",
        protected=False,
        default=False,
        created_at=datetime(2023, 1, 1),
        updated_at=datetime(2023, 1, 1),
    )


@pytest.fixture
def mock_pull_request(mock_user, mock_label, mock_comment) -> PullRequest:
    return PullRequest(
        number=1,
        title="Test PR",
        description="Test description",
        source_branch="feature-branch",
        target_branch="main",
        state="open",
        created_at=datetime(2023, 1, 1),
        updated_at=datetime(2023, 1, 1),
        merged_at=None,
        closed_at=None,
        author=mock_user,
        assignees=[mock_user],
        labels=[mock_label],
        comments=[mock_comment],
    )



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any, ClassVar

from githarbor.core.base import BaseOwner, BaseRepository
from githarbor.exceptions import AuthenticationError, ResourceNotFoundError
from githarbor.providers.aiogithubapi_provider import utils as aiogithubapitools
from githarbor.repositories import create_repository


if TYPE_CHECKING:
    from githarbor.core.models import User

logger = logging.getLogger(__name__)


class AioGitHubOwner(BaseOwner):
    """Owner implementation using aiogithubapi."""

    url_patterns: ClassVar[list[str]] = ["github.com"]
    is_async = True

    def __init__(self, username: str, token: str | None = None) -> None:
        """Initialize GitHub API client.

        Args:
            username: Username of the owner
            token: GitHub access token
        """
        from aiogithubapi import GitHubAPI
        from aiogithubapi.exceptions import GitHubAuthenticationException

        try:
            token = token or os.getenv("GITHUB_TOKEN")
            if not token:
                msg = "GitHub token is required"
                raise ValueError(msg)
            self._gh = GitHubAPI(token=token)
            self._name = username

        except GitHubAuthenticationException as e:
            msg = f"GitHub authentication failed: {e}"
            raise AuthenticationError(msg) from e

    @property
    def name(self) -> str:
        """The name of the owner."""
        return self._name

    @classmethod
    def from_url(cls, url: str, **kwargs: Any) -> AioGitHubOwner:
        """Create instance from URL.

        Args:
            url: URL like "https://github.com/octocat"
            **kwargs: Additional arguments including token
        """
        # Remove trailing slashes and handle both user & org URLs
        url = url.rstrip("/")
        username = url.split("/")[-1]
        return cls(username=username, token=kwargs.get("token"))

    async def list_repositories_async(self) -> list[BaseRepository]:
        """List all repositories owned by this user."""
        from aiogithubapi.exceptions import GitHubAuthenticationException

        try:
            response = await self._gh.user.repos()
            assert response.data
            return [
                create_repository(repo.html_url)
                for repo in response.data
                if repo.html_url
            ]
        except GitHubAuthenticationException as e:
            msg = f"Failed to list repositories: {e}"
            raise ResourceNotFoundError(msg) from e

    async def create_repository_async(
        self,
        name: str,
        description: str = "",
        private: bool = False,
    ) -> BaseRepository:
        """Create a new repository.

        Args:
            name: Repository name
            description: Repository description
            private: Whether to create a private repository

        Returns:
            Newly created repository
        """
        from aiogithubapi import GitHubRequestKwarg
        from aiogithubapi.exceptions import GitHubAuthenticationException
        from aiogithubapi.github import HttpMethod

        try:
            data = {
                "name": name,
                "description": description,
                "private": private,
            }
            kwargs: dict[Any, Any] = {GitHubRequestKwarg.METHOD: HttpMethod.POST}
            response = await self._gh.generic(
                endpoint="/user/repos",
                data=data,
                **kwargs,
            )
            assert response.data
            return create_repository(response.data["html_url"])
        except GitHubAuthenticationException as e:
            msg = f"Failed to create repository: {e}"
            raise ResourceNotFoundError(msg) from e

    async def get_user_async(self) -> User:
        """Get user information."""
        from aiogithubapi.exceptions import GitHubAuthenticationException

        try:
            response = await self._gh.users.get(self._name)
            assert response.data
            return aiogithubapitools.create_user_model(response.data)
        except GitHubAuthenticationException as e:
            msg = f"Failed to get user info: {e}"
            raise ResourceNotFoundError(msg) from e

    async def delete_repository_async(self, name: str) -> None:
        """Delete a repository.

        Args:
            name: Name of repository to delete
        """
        from aiogithubapi import GitHubRequestKwarg
        from aiogithubapi.exceptions import GitHubAuthenticationException
        from aiogithubapi.github import HttpMethod

        try:
            kwargs = {GitHubRequestKwarg.METHOD: HttpMethod.DELETE}
            await self._gh.generic(
                endpoint=f"/repos/{self._name}/{name}",
                **kwargs,  # type: ignore
            )
        except GitHubAuthenticationException as e:
            msg = f"Failed to delete repository: {e}"
            raise ResourceNotFoundError(msg) from e



"""Tools for converting aiogithubapi models to githarbor models."""

from __future__ import annotations

from datetime import datetime
import logging
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar, overload

from aiogithubapi.exceptions import GitHubException
import aiohttp
import upath

from githarbor.core.models import (
    Issue,
    Label,
    PullRequest,
    Release,
    Tag,
    User,
)
from githarbor.exceptions import ResourceNotFoundError


if TYPE_CHECKING:
    import os

    from aiogithubapi import GitHubLabelModel
    from aiogithubapi.models.issue import GitHubIssueModel
    from aiogithubapi.models.pull_request import GitHubPullRequestModel
    from aiogithubapi.models.release import GitHubReleaseModel
    from aiogithubapi.models.tag import GitHubTagModel
    from aiogithubapi.models.user import GitHubBaseUserModel, GitHubUserModel


logger = logging.getLogger(__name__)
T = TypeVar("T")
P = ParamSpec("P")


def parse_datetime(date_str: str | None) -> datetime | None:
    """Parse GitHub date string to datetime."""
    if not date_str:
        return None
    return datetime.fromisoformat(date_str.replace("Z", "+00:00"))


@overload
def create_user_model(user: None) -> None: ...


@overload
def create_user_model(user: GitHubBaseUserModel | GitHubUserModel) -> User: ...


def create_user_model(user: GitHubBaseUserModel | GitHubUserModel | None) -> User | None:
    """Convert aiogithubapi user model to GitHarbor user model."""
    if not user:
        return None

    data = {
        "username": user.login or "",
        "name": getattr(user, "name", None),
        "email": getattr(user, "email", None),
        "avatar_url": user.avatar_url,
        "url": user.html_url,
        "created_at": parse_datetime(getattr(user, "created_at", None)),
        "bio": getattr(user, "bio", None),
        "location": getattr(user, "location", None),
        "company": getattr(user, "company", None),
        "blog": getattr(user, "blog", None),
        "twitter_username": getattr(user, "twitter_username", None),
        "hireable": getattr(user, "hireable", None),
        "gravatar_id": user.gravatar_id,
        "followers": getattr(user, "followers", None),
        "following": getattr(user, "following", None),
        "public_repos": getattr(user, "public_repos", None),
    }
    return User(**data)  # pyright: ignore


def create_label_model(label: GitHubLabelModel) -> Label:
    """Convert aiogithubapi label model to GitHarbor label model."""
    return Label(
        name=label.name or "",
        color=label.color or "",
        description=label.description or "",
        url=label.url,
    )


def create_issue_model(issue: GitHubIssueModel) -> Issue:
    """Convert aiogithubapi issue model to GitHarbor issue model."""
    return Issue(
        number=int(issue.number or 0),
        title=str(issue.title or ""),
        description=str(issue.body or ""),
        state=str(issue.state or ""),
        created_at=parse_datetime(issue.created_at),
        updated_at=parse_datetime(issue.updated_at),
        closed_at=parse_datetime(issue.closed_at),
        closed=issue.state == "closed",
        author=create_user_model(issue.user),
        assignee=create_user_model(issue.assignee),
        labels=[create_label_model(lbl) for lbl in (issue.labels or [])],
        comments_count=int(issue.comments or 0),
        url=str(issue.html_url or ""),
        milestone=issue.milestone.title if issue.milestone else None,
    )


def create_pull_request_model(pr: GitHubPullRequestModel) -> PullRequest:
    """Convert aiogithubapi pull request model to GitHarbor pull request model."""
    return PullRequest(
        number=int(pr.number or 0),
        title=str(pr.title or ""),
        description=str(pr.body or ""),
        state=str(pr.state or ""),
        source_branch=str(pr.head.get("ref", "") if pr.head else ""),
        target_branch=str(pr.base.get("ref", "") if pr.base else ""),
        created_at=parse_datetime(pr.created_at),
        updated_at=parse_datetime(pr.updated_at),
        merged_at=parse_datetime(pr.merged_at),
        closed_at=parse_datetime(pr.closed_at),
        author=create_user_model(pr.user),
        assignees=[create_user_model(a) for a in (pr.assignees or [])],
        labels=[create_label_model(lbl) for lbl in (pr.labels or [])],
        merged_by=None,  # Not available
        review_comments_count=0,  # Not available
        commits_count=0,  # Not available
        additions=0,  # Not available
        deletions=0,  # Not available
        changed_files=0,  # Not available
        url=str(pr.html_url or ""),
    )


def _process_release_asset(asset: dict[str, Any]) -> dict[str, Any]:
    """Convert release asset data."""
    return {
        "name": str(asset["name"]),
        "url": str(asset["browser_download_url"]),
        "size": int(asset["size"]),
        "download_count": int(asset["download_count"]),
    }


def create_release_model(release: GitHubReleaseModel) -> Release:
    """Convert aiogithubapi release model to GitHarbor release model."""
    return Release(
        tag_name=str(release.tag_name or ""),
        name=str(release.name or release.tag_name or ""),
        description=str(release.body or ""),
        created_at=parse_datetime(release.created_at),
        published_at=parse_datetime(release.published_at),
        draft=bool(release.draft),
        prerelease=bool(release.prerelease),
        author=create_user_model(release.author),
        assets=[_process_release_asset(asset) for asset in (release.assets or [])],
        url=str(release.html_url or ""),
        target_commitish=str(release.target_commitish or ""),
    )


def create_tag_model(tag: GitHubTagModel) -> Tag:
    """Convert aiogithubapi tag model to GitHarbor tag model."""
    return Tag(
        name=str(tag.name or ""),
        sha=str(tag.commit.sha if tag.commit else ""),
        message=None,  # Not available
        created_at=None,  # Not available
        author=None,  # Not available
        url=None,  # Not available
    )


async def download_from_github(
    repository: str,
    path: str | os.PathLike[str],
    destination: str | os.PathLike[str],
    token: str | None = None,
    recursive: bool = False,
) -> None:
    """Download files from GitHub repository."""
    owner, repo = repository.split("/")

    dest = upath.UPath(destination)
    dest.mkdir(exist_ok=True, parents=True)

    url = f"https://raw.githubusercontent.com/{owner}/{repo}/main/{path}"
    headers = {"Authorization": f"token {token}"} if token else {}

    async with aiohttp.ClientSession() as session:
        try:
            async with session.get(url, headers=headers) as response:
                if response.status == 404:  # noqa: PLR2004
                    msg = f"Path not found: {path}"
                    raise ResourceNotFoundError(msg)
                content = await response.read()
                file_dest = dest / upath.UPath(path).name
                file_dest.write_bytes(content)
        except aiohttp.ClientError as e:
            msg = f"Failed to download {path}: {e}"
            raise GitHubException(msg) from e



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal


@dataclass
class FileChange:
    """Represents a file change in a diff."""

    path: str
    content: str | None  # None means file deletion
    mode: Literal["add", "modify", "delete"]
    old_path: str | None = None  # For renamed files


def parse_diff(diff_str: str) -> list[FileChange]:
    """Parse a unified diff string into a list of file changes.

    Uses the unidiff library for robust diff parsing.

    Args:
        diff_str: Unified diff string

    Returns:
        List of FileChange objects
    """
    import unidiff

    patch_set = unidiff.PatchSet(diff_str.splitlines(keepends=True))
    changes: list[FileChange] = []

    for patched_file in patch_set:
        if patched_file.is_rename:
            # Handle renamed files
            changes.append(
                FileChange(
                    path=patched_file.target_file,
                    old_path=patched_file.source_file,
                    content=patched_file.target_file[1:],  # Remove leading /
                    mode="modify",
                )
            )
            continue

        # Determine the change type
        if patched_file.is_added_file:
            mode = "add"
        elif patched_file.is_removed_file:
            mode = "delete"
        else:
            mode = "modify"

        # For deletions, we don't need content
        if mode == "delete":
            change = FileChange(path=patched_file.path, content=None, mode="delete")
            changes.append(change)
            continue
        # Reconstruct the final content
        lines = [ln.value for hunk in patched_file for ln in hunk if not ln.is_removed]
        change = FileChange(patched_file.path, content="".join(lines), mode=mode)  # type: ignore
        changes.append(change)

    return changes



"""Base repository class. All methods raise FeatureNotSupportedError by default."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Literal

from githarbor.exceptions import FeatureNotSupportedError


if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime
    import os

    from githarbor.core.models import (
        Branch,
        Commit,
        Issue,
        PullRequest,
        Release,
        Tag,
        User,
        Workflow,
        WorkflowRun,
    )

IssueState = Literal["open", "closed", "all"]
PullRequestState = Literal["open", "closed", "all"]


class BaseRepository:
    """Base repository class. All methods raise FeatureNotSupportedError by default."""

    is_async: ClassVar[bool] = False
    url_patterns: ClassVar[list[str]] = []
    _owner: str = ""
    _name: str = ""

    @property
    def name(self) -> str:
        """The name of the repository."""
        return self._name

    @property
    def owner(self) -> str:
        """The owner of the repository."""
        return self._owner

    @property
    def default_branch(self) -> str:
        """The default branch of this repository."""
        raise NotImplementedError

    @property
    def edit_base_uri(self) -> str | None:
        """The edit uri prefix of a repository."""
        return None

    @classmethod
    def supports_url(cls, url: str) -> bool:
        return any(pattern in url for pattern in cls.url_patterns)

    @classmethod
    def from_url(cls, url: str, **kwargs: Any) -> BaseRepository:
        msg = f"{cls.__name__} does not implement from_url"
        raise FeatureNotSupportedError(msg)

    def get_repo_user(self) -> User:
        msg = f"{self.__class__.__name__} does not implement get_repo_user"
        raise FeatureNotSupportedError(msg)

    def get_branch(self, name: str) -> Branch:
        msg = f"{self.__class__.__name__} does not implement get_branch"
        raise FeatureNotSupportedError(msg)

    def list_branches(self) -> list[Branch]:
        msg = f"{self.__class__.__name__} does not implement list_branches"
        raise FeatureNotSupportedError(msg)

    def get_pull_request(self, number: int) -> PullRequest:
        msg = f"{self.__class__.__name__} does not implement get_pull_request"
        raise FeatureNotSupportedError(msg)

    def list_pull_requests(self, state: PullRequestState = "open") -> list[PullRequest]:
        msg = f"{self.__class__.__name__} does not implement list_pull_requests"
        raise FeatureNotSupportedError(msg)

    def get_issue(self, issue_id: int) -> Issue:
        msg = f"{self.__class__.__name__} does not implement get_issue"
        raise FeatureNotSupportedError(msg)

    def list_issues(self, state: IssueState = "open") -> list[Issue]:
        msg = f"{self.__class__.__name__} does not implement list_issues"
        raise FeatureNotSupportedError(msg)

    def get_commit(self, sha: str) -> Commit:
        msg = f"{self.__class__.__name__} does not implement get_commit"
        raise FeatureNotSupportedError(msg)

    def list_commits(
        self,
        branch: str | None = None,
        since: datetime | None = None,
        until: datetime | None = None,
        author: str | None = None,
        path: str | None = None,
        max_results: int | None = None,
    ) -> list[Commit]:
        msg = f"{self.__class__.__name__} does not implement list_commits"
        raise FeatureNotSupportedError(msg)

    def get_workflow(self, workflow_id: str) -> Workflow:
        msg = f"{self.__class__.__name__} does not implement get_workflow"
        raise FeatureNotSupportedError(msg)

    def list_workflows(self) -> list[Workflow]:
        msg = f"{self.__class__.__name__} does not implement list_workflows"
        raise FeatureNotSupportedError(msg)

    def get_workflow_run(self, run_id: str) -> WorkflowRun:
        msg = f"{self.__class__.__name__} does not implement get_workflow_run"
        raise FeatureNotSupportedError(msg)

    def download(
        self,
        path: str | os.PathLike[str],
        destination: str | os.PathLike[str],
        recursive: bool = False,
    ) -> None:
        msg = f"{self.__class__.__name__} does not implement download"
        raise FeatureNotSupportedError(msg)

    def search_commits(
        self,
        query: str,
        branch: str | None = None,
        path: str | None = None,
        max_results: int | None = None,
    ) -> list[Commit]:
        msg = f"{self.__class__.__name__} does not implement search_commits"
        raise FeatureNotSupportedError(msg)

    def iter_files(
        self,
        path: str = "",
        ref: str | None = None,
        pattern: str | None = None,
    ) -> Iterator[str]:
        msg = f"{self.__class__.__name__} does not implement iter_files"
        raise FeatureNotSupportedError(msg)

    def get_contributors(
        self,
        sort_by: Literal["commits", "name", "date"] = "commits",
        limit: int | None = None,
    ) -> list[User]:
        msg = f"{self.__class__.__name__} does not implement get_contributors"
        raise FeatureNotSupportedError(msg)

    def get_languages(self) -> dict[str, int]:
        msg = f"{self.__class__.__name__} does not implement get_languages"
        raise FeatureNotSupportedError(msg)

    def compare_branches(
        self,
        base: str,
        head: str,
        include_commits: bool = True,
        include_files: bool = True,
        include_stats: bool = True,
    ) -> dict[str, Any]:
        msg = f"{self.__class__.__name__} does not implement compare_branches"
        raise FeatureNotSupportedError(msg)

    def get_latest_release(
        self,
        include_drafts: bool = False,
        include_prereleases: bool = False,
    ) -> Release:
        msg = f"{self.__class__.__name__} does not implement get_latest_release"
        raise FeatureNotSupportedError(msg)

    def list_releases(
        self,
        include_drafts: bool = False,
        include_prereleases: bool = False,
        limit: int | None = None,
    ) -> list[Release]:
        msg = f"{self.__class__.__name__} does not implement list_releases"
        raise FeatureNotSupportedError(msg)

    def get_release(self, tag: str) -> Release:
        msg = f"{self.__class__.__name__} does not implement get_release"
        raise FeatureNotSupportedError(msg)

    def get_tag(self, name: str) -> Tag:
        msg = f"{self.__class__.__name__} does not implement get_tag"
        raise FeatureNotSupportedError(msg)

    def list_tags(self) -> list[Tag]:
        msg = f"{self.__class__.__name__} does not implement list_tags"
        raise FeatureNotSupportedError(msg)

    def create_pull_request(
        self,
        title: str,
        body: str,
        head_branch: str,
        base_branch: str,
        draft: bool = False,
    ) -> PullRequest:
        msg = f"{self.__class__.__name__} does not implement create_pull_request"
        raise FeatureNotSupportedError(msg)

    def create_branch(
        self,
        name: str,
        base_commit: str,
    ) -> Branch:
        """Create a new branch at the specified commit."""
        msg = f"{self.__class__.__name__} does not implement create_branch"
        raise FeatureNotSupportedError(msg)

    def create_pull_request_from_diff(
        self,
        title: str,
        body: str,
        base_branch: str,
        diff: str,
        head_branch: str | None = None,
        draft: bool = False,
    ) -> PullRequest:
        """Create a pull request from a diff string."""
        msg = (
            f"{self.__class__.__name__} does not implement create_pull_request_from_diff"
        )
        raise FeatureNotSupportedError(msg)

    async def get_repo_user_async(self) -> User:
        """Get repository owner information asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_repo_user_async"
        raise FeatureNotSupportedError(msg)

    async def get_branch_async(self, name: str) -> Branch:
        """Get branch information asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_branch_async"
        raise FeatureNotSupportedError(msg)

    async def get_pull_request_async(self, number: int) -> PullRequest:
        """Get pull request information asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_pull_request_async"
        raise FeatureNotSupportedError(msg)

    async def list_pull_requests_async(
        self,
        state: PullRequestState = "open",
    ) -> list[PullRequest]:
        """List pull requests asynchronously."""
        msg = f"{self.__class__.__name__} does not implement list_pull_requests_async"
        raise FeatureNotSupportedError(msg)

    async def get_issue_async(self, issue_id: int) -> Issue:
        """Get issue information asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_issue_async"
        raise FeatureNotSupportedError(msg)

    async def list_issues_async(self, state: IssueState = "open") -> list[Issue]:
        """List issues asynchronously."""
        msg = f"{self.__class__.__name__} does not implement list_issues_async"
        raise FeatureNotSupportedError(msg)

    async def get_commit_async(self, sha: str) -> Commit:
        """Get commit information asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_commit_async"
        raise FeatureNotSupportedError(msg)

    async def list_commits_async(
        self,
        branch: str | None = None,
        since: datetime | None = None,
        until: datetime | None = None,
        author: str | None = None,
        path: str | None = None,
        max_results: int | None = None,
    ) -> list[Commit]:
        """List commits asynchronously."""
        msg = f"{self.__class__.__name__} does not implement list_commits_async"
        raise FeatureNotSupportedError(msg)

    async def get_workflow_async(self, workflow_id: str) -> Workflow:
        """Get workflow information asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_workflow_async"
        raise FeatureNotSupportedError(msg)

    async def list_workflows_async(self) -> list[Workflow]:
        """List workflows asynchronously."""
        msg = f"{self.__class__.__name__} does not implement list_workflows_async"
        raise FeatureNotSupportedError(msg)

    async def get_workflow_run_async(self, run_id: str) -> WorkflowRun:
        """Get workflow run information asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_workflow_run_async"
        raise FeatureNotSupportedError(msg)

    async def download_async(
        self,
        path: str | os.PathLike[str],
        destination: str | os.PathLike[str],
        recursive: bool = False,
    ) -> None:
        """Download repository content asynchronously."""
        msg = f"{self.__class__.__name__} does not implement download_async"
        raise FeatureNotSupportedError(msg)

    async def search_commits_async(
        self,
        query: str,
        branch: str | None = None,
        path: str | None = None,
        max_results: int | None = None,
    ) -> list[Commit]:
        """Search commits asynchronously."""
        msg = f"{self.__class__.__name__} does not implement search_commits_async"
        raise FeatureNotSupportedError(msg)

    async def get_contributors_async(
        self,
        sort_by: Literal["commits", "name", "date"] = "commits",
        limit: int | None = None,
    ) -> list[User]:
        """Get repository contributors asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_contributors_async"
        raise FeatureNotSupportedError(msg)

    async def get_languages_async(self) -> dict[str, int]:
        """Get repository language statistics asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_languages_async"
        raise FeatureNotSupportedError(msg)

    async def compare_branches_async(
        self,
        base: str,
        head: str,
        include_commits: bool = True,
        include_files: bool = True,
        include_stats: bool = True,
    ) -> dict[str, Any]:
        """Compare two branches asynchronously."""
        msg = f"{self.__class__.__name__} does not implement compare_branches_async"
        raise FeatureNotSupportedError(msg)

    async def get_latest_release_async(
        self,
        include_drafts: bool = False,
        include_prereleases: bool = False,
    ) -> Release:
        """Get latest release asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_latest_release_async"
        raise FeatureNotSupportedError(msg)

    async def list_releases_async(
        self,
        include_drafts: bool = False,
        include_prereleases: bool = False,
        limit: int | None = None,
    ) -> list[Release]:
        """List releases asynchronously."""
        msg = f"{self.__class__.__name__} does not implement list_releases_async"
        raise FeatureNotSupportedError(msg)

    async def get_release_async(self, tag: str) -> Release:
        """Get release by tag asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_release_async"
        raise FeatureNotSupportedError(msg)

    async def get_tag_async(self, name: str) -> Tag:
        """Get tag information asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_tag_async"
        raise FeatureNotSupportedError(msg)

    async def list_tags_async(self) -> list[Tag]:
        """List all tags asynchronously."""
        msg = f"{self.__class__.__name__} does not implement list_tags_async"
        raise FeatureNotSupportedError(msg)

    async def create_pull_request_async(
        self,
        title: str,
        body: str,
        head_branch: str,
        base_branch: str,
        draft: bool = False,
    ) -> PullRequest:
        msg = f"{self.__class__.__name__} does not implement create_pull_request_async"
        raise FeatureNotSupportedError(msg)

    async def create_branch_async(
        self,
        name: str,
        base_commit: str,
    ) -> Branch:
        """Create a new branch at the specified commit asynchronously."""
        msg = f"{self.__class__.__name__} does not implement create_branch_async"
        raise FeatureNotSupportedError(msg)

    async def create_pull_request_from_diff_async(
        self,
        title: str,
        body: str,
        base_branch: str,
        diff: str,
        head_branch: str | None = None,
        draft: bool = False,
    ) -> PullRequest:
        """Create a pull request from a diff string asynchronously."""
        fn_name = "create_pull_request_from_diff_async"
        msg = f"{self.__class__.__name__} does not implement {fn_name}"
        raise FeatureNotSupportedError(msg)


class BaseOwner:
    """Base class for repository owners."""

    is_async: ClassVar[bool] = False
    url_patterns: ClassVar[list[str]] = []
    _name: str = ""

    def list_repositories(self) -> list[BaseRepository]:
        msg = f"{self.__class__.__name__} does not implement list_repositories"
        raise FeatureNotSupportedError(msg)

    @property
    def name(self) -> str:
        """The name of the repository."""
        return self._name

    def create_repository(
        self,
        name: str,
        description: str = "",
        private: bool = False,
    ) -> BaseRepository:
        msg = f"{self.__class__.__name__} does not implement create_repository"
        raise FeatureNotSupportedError(msg)

    def get_user(self) -> User:
        msg = f"{self.__class__.__name__} does not implement get_user"
        raise FeatureNotSupportedError(msg)

    def delete_repository(self, name: str) -> None:
        msg = f"{self.__class__.__name__} does not implement delete_repository"
        raise FeatureNotSupportedError(msg)

    async def list_repositories_async(self) -> list[BaseRepository]:
        """List repositories asynchronously."""
        msg = f"{self.__class__.__name__} does not implement list_repositories_async"
        raise FeatureNotSupportedError(msg)

    async def create_repository_async(
        self,
        name: str,
        description: str = "",
        private: bool = False,
    ) -> BaseRepository:
        """Create repository asynchronously."""
        msg = f"{self.__class__.__name__} does not implement create_repository_async"
        raise FeatureNotSupportedError(msg)

    async def get_user_async(self) -> User:
        """Get user information asynchronously."""
        msg = f"{self.__class__.__name__} does not implement get_user_async"
        raise FeatureNotSupportedError(msg)

    async def delete_repository_async(self, name: str) -> None:
        """Delete repository asynchronously."""
        msg = f"{self.__class__.__name__} does not implement delete_repository_async"
        raise FeatureNotSupportedError(msg)

    @classmethod
    def supports_url(cls, url: str) -> bool:
        return any(pattern in url for pattern in cls.url_patterns)

    @classmethod
    def from_url(cls, url: str, **kwargs: Any) -> BaseOwner:
        msg = f"{cls.__name__} does not implement from_url"
        raise FeatureNotSupportedError(msg)



"""Azure DevOps repository implementation."""

from __future__ import annotations

import fnmatch
import os
from typing import TYPE_CHECKING, Any, ClassVar
from urllib.parse import urlparse

from githarbor.core.base import BaseRepository, IssueState, PullRequestState
from githarbor.core.models import (
    Branch,
    Commit,
    Issue,
    PullRequest,
    Release,
    User,
)
from githarbor.exceptions import AuthenticationError, ResourceNotFoundError
from githarbor.providers.azure_provider import utils as azuretools


if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

    from azure.devops.v7_1.build.build_client import BuildClient
    from azure.devops.v7_1.git.git_client import GitClient
    from azure.devops.v7_1.git.models import GitCommit, GitPullRequest, GitRepository
    from azure.devops.v7_1.work_item_tracking.work_item_tracking_client import (
        WorkItemTrackingClient,
    )


class AzureRepository(BaseRepository):
    """Azure DevOps repository implementation."""

    url_patterns: ClassVar[list[str]] = ["dev.azure.com", "visualstudio.com"]

    def __init__(
        self,
        organization: str,
        project: str,
        name: str,
        token: str | None = None,
    ) -> None:
        """Initialize Azure DevOps repository.

        Args:
            organization: Azure DevOps organization name
            project: Project name
            name: Repository name
            token: Personal access token for authentication

        Raises:
            AuthenticationError: If authentication fails
            ValueError: If token is missing
        """
        from azure.devops.connection import Connection
        from msrest.authentication import BasicAuthentication

        t = token or os.getenv("AZURE_DEVOPS_PAT")
        if not t:
            msg = "Azure DevOps PAT token is required"
            raise ValueError(msg)
        try:
            credentials = BasicAuthentication("", t)
            organization_url = f"https://dev.azure.com/{organization}"
            self._connection = Connection(base_url=organization_url, creds=credentials)

            self._git_client: GitClient = self._connection.clients.get_git_client()
            self._work_client: WorkItemTrackingClient = (
                self._connection.clients.get_work_item_tracking_client()
            )
            self._build_client: BuildClient = self._connection.clients.get_build_client()

            self._project = project
            self._name = name
            self._owner = organization
            self._repo: GitRepository = self._git_client.get_repository(
                name,
                project=project,
            )

        except Exception as e:
            msg = f"Azure DevOps authentication failed: {e!s}"
            raise AuthenticationError(msg) from e

    @classmethod
    def from_url(cls, url: str, **kwargs: Any) -> AzureRepository:
        """Create from URL.

        Example URL: 'https://dev.azure.com/org/project/_git/repo'
        """
        parsed = urlparse(url)
        parts = parsed.path.strip("/").split("/")

        if len(parts) < 4:  # noqa: PLR2004
            msg = f"Invalid Azure DevOps URL: {url}"
            raise ValueError(msg)

        organization = parts[0]
        project = parts[1]
        repo_name = parts[3]  # After '_git'

        return cls(
            organization=organization,
            project=project,
            name=repo_name,
            token=kwargs.get("token"),
        )

    @property
    def default_branch(self) -> str:
        """Default branch name."""
        return self._repo.default_branch or "main"

    @azuretools.handle_azure_errors("Failed to get user {username}")
    def get_user(self, username: str | None = None) -> User:
        """Get user information.

        If username is not provided, returns the authenticated user.

        Args:
            username: Optional username to get information for.

        Returns:
            User model with user information.
        """
        # Get the identity client
        identity_client = self._connection.clients.get_identity_client()

        if username is None:
            # Get the authenticated user
            core_client = self._connection.clients.get_core_client()
            user_info = core_client.get_authorized_user()
        else:
            # Search for the specific user
            user_descriptor = identity_client.read_identities(
                search_filter=f"General,localAccount,{username}",
            )
            if not user_descriptor or not user_descriptor[0]:
                msg = f"User {username} not found"
                raise ResourceNotFoundError(msg)
            user_info = user_descriptor[0]

        return User(
            username=user_info.properties.get("Account", user_info.unique_name),
            name=user_info.display_name or user_info.unique_name,
            email=user_info.properties.get("Mail", ""),
            avatar_url=user_info.properties.get("Avatar", ""),
        )

    @azuretools.handle_azure_errors("Failed to get branch {name}")
    def get_branch(self, name: str) -> Branch:
        """Get branch information."""
        branch = self._git_client.get_branch(
            repository_id=self._repo.id,
            name=name,
            project=self._project,
        )
        commit = self._git_client.get_commit(branch.commit.commit_id, self._repo.id)
        return Branch(
            name=branch.name,
            sha=branch.commit.commit_id,
            protected=False,  # Azure DevOps handles branch protection differently
            default=branch.name == self.default_branch,
            created_at=None,  # Not provided by Azure API
            updated_at=None,  # Not provided by Azure API
            last_commit_date=commit.author.date,
            last_commit_message=commit.comment,
            last_commit_author=azuretools.create_user_model(commit.author),
        )

    @azuretools.handle_azure_errors("Failed to get pull request {number}")
    def get_pull_request(self, number: int) -> PullRequest:
        """Get pull request by number."""
        pr: GitPullRequest = self._git_client.get_pull_request_by_id(
            number,
            self._project,
        )
        return azuretools.create_pull_request_model(pr)

    @azuretools.handle_azure_errors("Failed to list pull requests")
    def list_pull_requests(self, state: PullRequestState = "open") -> list[PullRequest]:
        """List pull requests."""
        # Map state to Azure DevOps status
        from azure.devops.v7_1.git.models import GitPullRequestSearchCriteria

        status_map = {"open": "active", "closed": "completed", "all": "all"}
        azure_status = status_map.get(state, "active")
        criteria = GitPullRequestSearchCriteria(status=azure_status)
        prs = self._git_client.get_pull_requests(
            self._repo.id, search_criteria=criteria, project=self._project
        )
        return [azuretools.create_pull_request_model(pr) for pr in prs]

    @azuretools.handle_azure_errors("Failed to get issue {issue_id}")
    def get_issue(self, issue_id: int) -> Issue:
        """Get issue by ID (work item in Azure DevOps)."""
        work_item = self._work_client.get_work_item(issue_id, self._project)
        return azuretools.create_issue_model(work_item)

    @azuretools.handle_azure_errors("Failed to list issues")
    def list_issues(self, state: IssueState = "open") -> list[Issue]:
        """List issues (work items in Azure DevOps)."""
        # Build WIQL query based on state
        state_map = {"open": "Active", "closed": "Closed", "all": ""}
        azure_state = state_map.get(state, "Active")
        state_clause = f"AND [System.State] = '{azure_state}'" if azure_state else ""
        wiql = (
            f"SELECT [System.Id] FROM WorkItems "
            f"WHERE [System.TeamProject] = '{self._project}' "
            f"{state_clause} ORDER BY [System.ChangedDate] DESC"
        )
        wiql_result = self._work_client.query_by_wiql({"query": wiql})
        if not wiql_result.work_items:
            return []

        work_items = self._work_client.get_work_items(
            [wi.id for wi in wiql_result.work_items],
        )
        return [azuretools.create_issue_model(wi) for wi in work_items]

    @azuretools.handle_azure_errors("Failed to get commit {sha}")
    def get_commit(self, sha: str) -> Commit:
        """Get commit by SHA."""
        commit: GitCommit = self._git_client.get_commit(
            commit_id=sha,
            repository_id=self._repo.id,
            project=self._project,
        )
        return azuretools.create_commit_model(commit)

    @azuretools.handle_azure_errors("Failed to list commits")
    def list_commits(
        self,
        branch: str | None = None,
        since: datetime | None = None,
        until: datetime | None = None,
        author: str | None = None,
        path: str | None = None,
        max_results: int | None = None,
    ) -> list[Commit]:
        """List repository commits with optional filters."""
        commits = self._git_client.get_commits(
            repository_id=self._repo.id,
            search_criteria=None,
            project=self._project,
            # branch_name=branch,
            # from_date=since,
            # to_date=until,
            # author=author,
            # item_path=path,
            top=max_results,
        )
        return [azuretools.create_commit_model(c) for c in commits]

    @azuretools.handle_azure_errors("Failed to download file {path}")
    def download(
        self,
        path: str | os.PathLike[str],
        destination: str | os.PathLike[str],
        recursive: bool = False,
    ) -> None:
        """Download file(s) from repository."""
        return azuretools.download_from_azure(
            organization=self._owner,
            project=self._project,
            repo=self._repo.id,
            path=path,
            destination=destination,
            recursive=recursive,
        )

    @azuretools.handle_azure_errors("Failed to list files for {path}")
    def iter_files(
        self,
        path: str = "",
        ref: str | None = None,
        pattern: str | None = None,
    ) -> Iterator[str]:
        """Iterate over repository files."""
        items = self._git_client.get_items(
            repository_id=self._repo.id,
            project=self._project,
            recursion_level="full",
            version_descriptor={"version": ref} if ref else None,
        )
        for item in items:
            if item.is_folder:
                continue
            if not pattern or fnmatch.fnmatch(item.path, pattern):
                yield item.path

    @azuretools.handle_azure_errors("Failed to get latest release")
    def get_latest_release(
        self,
        include_drafts: bool = False,
        include_prereleases: bool = False,
    ) -> Release:
        """Get latest release (mapped from Git tags)."""
        tags = self._git_client.get_tags(
            repository_id=self._repo.id,
            project=self._project,
        )
        if not tags:
            msg = "No releases found"
            raise ResourceNotFoundError(msg)

        latest_tag = sorted(tags, key=lambda t: t.commit.committer.date, reverse=True)[0]
        return azuretools.create_release_model(latest_tag)

    @azuretools.handle_azure_errors("Failed to create pull request")
    def create_pull_request(
        self,
        title: str,
        body: str,
        head_branch: str,
        base_branch: str,
        draft: bool = False,
    ) -> PullRequest:
        """Create a pull request."""
        from azure.devops.v7_1.git.models import GitPullRequest

        request = GitPullRequest(
            source_ref_name=f"refs/heads/{head_branch}",
            target_ref_name=f"refs/heads/{base_branch}",
            title=title,
            description=body,
            is_draft=draft,
        )
        pr = self._git_client.create_pull_request(
            request,
            repository_id=self._repo.id,
            project=self._project,
        )
        return azuretools.create_pull_request_model(pr)


if __name__ == "__main__":
    pass



"""Data types for Githarbor."""

from __future__ import annotations

import reprlib
from typing import TypeVar


T = TypeVar("T")


class NiceReprList(list[T]):
    def __repr__(self):
        return reprlib.repr(list(self))



"""Module containing local repository implementation."""

from __future__ import annotations

import fnmatch
import functools
import pathlib
from typing import TYPE_CHECKING, Any, ClassVar

from githarbor.core.base import BaseRepository
from githarbor.exceptions import ResourceNotFoundError
from githarbor.providers.local_provider import utils as localtools


if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime
    import os

    from githarbor.core.models import Branch, Commit, Tag
    from githarbor.core.proxy import Repository


class LocalRepository(BaseRepository):
    """Local Git repository implementation using GitPython."""

    url_patterns: ClassVar[list[str]] = []  # Local repos don't have URL patterns

    def __init__(self, path: str | os.PathLike[str]) -> None:
        import git

        try:
            self.path = pathlib.Path(path)
            self.repo = git.Repo(self.path)
            self._name = self.path.name
            self._owner = self.path.parent.name  # or None?
        except (git.InvalidGitRepositoryError, git.NoSuchPathError) as e:
            msg = f"Not a valid git repository: {path}"
            raise ResourceNotFoundError(msg) from e

    @classmethod
    def from_url(cls, url: str, **_: Any) -> LocalRepository:
        return cls(url)

    @classmethod
    def supports_url(cls, url: str) -> bool:
        return pathlib.Path(url).exists()

    @property
    def default_branch(self) -> str:
        return self.repo.active_branch.name

    @localtools.handle_git_errors("Failed to get branch {name}")
    def get_branch(self, name: str) -> Branch:
        branch = self.repo.heads[name]
        return localtools.create_branch_model(
            branch,
            is_default=branch.name == self.default_branch,
        )

    @localtools.handle_git_errors("Failed to list branches")
    def list_branches(self) -> list[Branch]:
        return [
            localtools.create_branch_model(
                branch,
                is_default=branch.name == self.default_branch,
            )
            for branch in self.repo.heads
        ]

    @localtools.handle_git_errors("Failed to get commit {sha}")
    def get_commit(self, sha: str) -> Commit:
        commit = self.repo.commit(sha)
        return localtools.create_commit_model(commit)

    @localtools.handle_git_errors("Failed to list commits")
    def list_commits(
        self,
        branch: str | None = None,
        since: datetime | None = None,
        until: datetime | None = None,
        author: str | None = None,
        path: str | None = None,
        max_results: int | None = None,
    ) -> list[Commit]:
        commits = self.repo.iter_commits(rev=branch or self.default_branch)
        filtered = localtools.filter_commits(
            commits,
            since=since,
            until=until,
            author=author,
            path=path,
            max_results=max_results,
        )
        return [localtools.create_commit_model(commit) for commit in filtered]

    @localtools.handle_git_errors("Failed to iterate files")
    def iter_files(
        self,
        path: str = "",
        ref: str | None = None,
        pattern: str | None = None,
    ) -> Iterator[str]:
        tree = self.repo.head.commit.tree if ref is None else self.repo.commit(ref).tree
        for blob in tree.traverse():
            # Skip dirs
            if blob.type != "blob":  # type: ignore
                continue
            file_path = str(blob.path)  # type: ignore
            if (not pattern or fnmatch.fnmatch(file_path, pattern)) and (
                not path or file_path.startswith(path)
            ):
                yield file_path

    @localtools.handle_git_errors("Failed to get tag {name}")
    def get_tag(self, name: str) -> Tag:
        tag = self.repo.tags[name]
        return localtools.create_tag_model(tag.tag, tag.commit)  # type: ignore

    @localtools.handle_git_errors("Failed to list tags")
    def list_tags(self) -> list[Tag]:
        return [
            localtools.create_tag_model(tag.tag, tag.commit)  # type: ignore
            for tag in self.repo.tags  # type: ignore
        ]

    @functools.cached_property
    def remote_repository(self) -> Repository:
        """Get the remote code repository."""
        from githarbor import create_repository

        return create_repository(self.repo.remotes.origin.url)


if __name__ == "__main__":
    repo = LocalRepository(".")
    for commit in repo.list_commits(max_results=5):
        print(f"{commit.sha[:8]}: {commit.message.splitlines()[0]}")
    print(repo.list_tags())



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, ParamSpec, TypeVar, overload

from githarbor.core.models import (
    Branch,
    Commit,
    Issue,
    Label,
    PullRequest,
    Release,
    Tag,
    User,
)
from githarbor.exceptions import ResourceNotFoundError


if TYPE_CHECKING:
    from collections.abc import Callable

    from giteapy.models import (
        Branch as GiteaBranch,
        Commit as GiteaCommit,
        Issue as GiteaIssue,
        Label as GiteaLabel,
        PullRequest as GiteaPullRequest,
        Release as GiteaRelease,
        Tag as GiteaTag,
        User as GiteaUser,
    )

T = TypeVar("T")
P = ParamSpec("P")


def handle_api_errors(error_msg: str) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator to handle Gitea API exceptions consistently.

    Args:
        error_msg: Base error message to use in exception
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            from giteapy.rest import ApiException

            try:
                return func(*args, **kwargs)
            except ApiException as e:
                msg = f"{error_msg}: {e!s}"
                raise ResourceNotFoundError(msg) from e

        return wrapper

    return decorator


@overload
def create_user_model(gitea_user: None) -> None: ...


@overload
def create_user_model(gitea_user: GiteaUser) -> User: ...


def create_user_model(gitea_user: GiteaUser | None) -> User | None:
    """Create User model from Gitea user object.

    Args:
        gitea_user: Gitea user object to convert

    Returns:
        Converted User model or None if input is None
    """
    if not gitea_user:
        return None

    return User(
        username=gitea_user.login,
        name=gitea_user.full_name,
        email=gitea_user.email,
        avatar_url=gitea_user.avatar_url,
        created_at=gitea_user.created,
        bio=gitea_user.description,
        location=gitea_user.location,
        is_admin=gitea_user.is_admin,
        last_activity_on=gitea_user.last_login,
        blog=gitea_user.website,
        url=gitea_user.html_url,
        followers=getattr(gitea_user, "followers_count", None),
        following=getattr(gitea_user, "following_count", None),
        public_repos=getattr(gitea_user, "public_repos", None),
        linkedin=getattr(gitea_user, "linkedin", None),
        skype=getattr(gitea_user, "skype", None),
    )


def create_label_model(gitea_label: GiteaLabel) -> Label:
    """Create Label model from Gitea label object."""
    return Label(
        name=gitea_label.name,
        color=gitea_label.color,
        description=gitea_label.description or "",
        url=gitea_label.url,
    )


def create_pull_request_model(pr: GiteaPullRequest) -> PullRequest:
    """Create PullRequest model from Gitea PR object."""
    return PullRequest(
        number=pr.number,
        title=pr.title,
        description=pr.body or "",
        state=pr.state,
        source_branch=pr.head.ref,
        target_branch=pr.base.ref,
        created_at=pr.created_at,
        updated_at=pr.updated_at,
        merged_at=pr.merged_at,
        closed_at=pr.closed_at,
        author=create_user_model(pr.user),
        labels=[create_label_model(label) for label in (pr.labels or [])],
        url=pr.html_url,
    )


def create_issue_model(issue: GiteaIssue) -> Issue:
    """Create Issue model from Gitea issue object."""
    return Issue(
        number=issue.number,
        title=issue.title,
        description=issue.body or "",
        state=issue.state,
        created_at=issue.created_at,
        updated_at=issue.updated_at,
        closed_at=issue.closed_at,
        closed=issue.state == "closed",
        author=create_user_model(issue.user),
        assignee=create_user_model(issue.assignee),
        labels=[create_label_model(label) for label in (issue.labels or [])],
        url=issue.html_url,
    )


def create_commit_model(commit: GiteaCommit) -> Commit:
    """Create Commit model from Gitea commit object."""
    return Commit(
        sha=commit.sha,
        message=commit.commit.message,
        created_at=commit.commit.author._date,
        author=User(
            username=commit.commit.author.name,
            email=commit.commit.author.email,
            name=commit.commit.author.name,
        ),
        url=commit.html_url,
    )


def create_release_model(release: GiteaRelease) -> Release:
    """Create Release model from Gitea release object."""
    return Release(
        tag_name=release.tag_name,
        name=release.name,
        description=release.body or "",
        created_at=release.created_at,
        published_at=release.published_at,
        draft=release.draft,
        prerelease=release.prerelease,
        author=create_user_model(release.author),
        assets=[
            {
                "name": asset.name,
                "url": asset.browser_download_url,
                "size": asset.size,
                "download_count": asset.download_count,
            }
            for asset in release.assets
        ],
        url=release.url,
        target_commitish=release.target_commitish,
    )


def create_tag_model(gitea_tag: GiteaTag) -> Tag:
    """Create Tag model from Gitea tag object."""
    return Tag(
        name=gitea_tag.name,
        sha=gitea_tag.id,  # Gitea uses 'id' for the SHA
        message=gitea_tag.message,
        created_at=getattr(gitea_tag, "created_at", None),
        author=create_user_model(gitea_tag.tagger)
        if hasattr(gitea_tag, "tagger")
        else None,
        url=getattr(gitea_tag, "url", None),
        verified=bool(getattr(gitea_tag, "verification", {}).get("verified", False)),
    )


def create_branch_model(gitea_branch: GiteaBranch) -> Branch:
    """Create Branch model from Gitea branch object."""
    return Branch(
        name=gitea_branch.name,
        sha=gitea_branch.commit.id,
        protected=gitea_branch.protected,
        created_at=None,  # Gitea API doesn't provide this
        updated_at=None,  # Gitea API doesn't provide this
    )



"""Gitea repository implementation."""

from __future__ import annotations

import fnmatch
import logging
import os
from typing import TYPE_CHECKING, Any, ClassVar, Literal
from urllib.parse import urlparse

import upath

from githarbor.core.base import BaseRepository, IssueState, PullRequestState
from githarbor.exceptions import AuthenticationError, ResourceNotFoundError
from githarbor.providers.gitea_provider import utils as giteatools


logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

    import giteapy

    from githarbor.core.models import (
        Branch,
        Commit,
        Issue,
        PullRequest,
        Release,
        Tag,
        User,
        Workflow,
        WorkflowRun,
    )


class GiteaRepository(BaseRepository):
    """Gitea repository implementation."""

    url_patterns: ClassVar[list[str]] = [
        "gitea.com",
        "codeberg.org",
    ]  # Add your Gitea instances here

    def __init__(
        self,
        owner: str,
        name: str,
        token: str | None = None,
        url: str = "https://gitea.com",
    ):
        import giteapy
        from giteapy.rest import ApiException

        t = token or os.getenv("GITEA_TOKEN")
        if not t:
            msg = "Gitea token is required"
            raise ValueError(msg)
        configuration = giteapy.Configuration()
        configuration.host = url.rstrip("/") + "/api/v1"
        configuration.api_key["token"] = t
        # configuration.temp_folder_path = ...
        self._owner = owner
        self._name = name
        try:
            self._api = giteapy.ApiClient(configuration)
            self._org_api = giteapy.OrganizationApi(self._api)
            self._repo_api = giteapy.RepositoryApi(self._api)
            self._issues_api = giteapy.IssueApi(self._api)
            self._user_api = giteapy.UserApi(self._api)
            # Verify access and get repo info
            self._repo: giteapy.RepositoryApi = self._repo_api.repo_get(owner, name)
        except ApiException as e:
            msg = f"Gitea authentication failed: {e!s}"
            raise AuthenticationError(msg) from e

    @classmethod
    def from_url(cls, url: str, **kwargs: Any) -> GiteaRepository:
        """Create from URL like 'https://gitea.com/owner/repo'."""
        parsed = urlparse(url)
        parts = parsed.path.strip("/").split("/")
        if len(parts) < 2:  # noqa: PLR2004
            msg = f"Invalid Gitea URL: {url}"
            raise ValueError(msg)

        return cls(
            owner=parts[0],
            name=parts[1],
            token=kwargs.get("token"),
            url=f"{parsed.scheme}://{parsed.netloc}",
        )

    @property
    def default_branch(self) -> str:
        return self._repo.default_branch

    @property
    def edit_base_uri(self):
        return f"_edit/{self.default_branch}/"

    @giteatools.handle_api_errors("Failed to get branch")
    def get_branch(self, name: str) -> Branch:
        """Get a specific branch by name."""
        branch = self._repo_api.repo_get_branch(self._owner, self._name, name)
        return giteatools.create_branch_model(branch)

    @giteatools.handle_api_errors("Failed to get repository owner info")
    def get_repo_user(self) -> User:
        """Get user (repository owner) information."""
        user = self._user_api.user_get_current()
        return giteatools.create_user_model(user)

    @giteatools.handle_api_errors("Failed to get pull request")
    def get_pull_request(self, number: int) -> PullRequest:
        """Get a specific pull request by number."""
        pr = self._repo_api.repo_get_pull(self._owner, self._name, number)
        return giteatools.create_pull_request_model(pr)

    @giteatools.handle_api_errors("Failed to list pull requests")
    def list_pull_requests(self, state: PullRequestState = "open") -> list[PullRequest]:
        """List pull requests."""
        prs = self._repo_api.repo_list_pull_requests(
            self._owner,
            self._name,
            state=state,
        )
        assert isinstance(prs, list)
        return [giteatools.create_pull_request_model(pr) for pr in prs]

    @giteatools.handle_api_errors("Failed to get issue")
    def get_issue(self, issue_id: int) -> Issue:
        """Get a specific issue by ID."""
        issue: giteapy.Issue = self._issues_api.issue_get_issue(
            self._owner, self._name, issue_id
        )
        return giteatools.create_issue_model(issue)

    @giteatools.handle_api_errors("Failed to list issues")
    def list_issues(self, state: IssueState = "open") -> list[Issue]:
        """List repository issues."""
        issues: list[giteapy.Issue] = self._issues_api.issue_list_issues(
            self._owner,
            self._name,
            state=state,
        )
        return [giteatools.create_issue_model(issue) for issue in issues]

    @giteatools.handle_api_errors("Failed to get commit")
    def get_commit(self, sha: str) -> Commit:
        """Get a specific commit by SHA."""
        commit: giteapy.Commit = self._repo_api.repo_get_single_commit(
            self._owner, self._name, sha
        )
        return giteatools.create_commit_model(commit)

    @giteatools.handle_api_errors("Failed to list commits")
    def list_commits(
        self,
        branch: str | None = None,
        since: datetime | None = None,
        until: datetime | None = None,
        author: str | None = None,
        path: str | None = None,
        max_results: int | None = None,
    ) -> list[Commit]:
        """List repository commits with optional filters."""
        kwargs: dict[str, Any] = {}
        if branch:
            kwargs["sha"] = branch
        if since:
            kwargs["since"] = since.isoformat()
        if until:
            kwargs["until"] = until.isoformat()
        if path:
            kwargs["path"] = path
        if max_results:
            kwargs["limit"] = max_results

        commits = self._repo_api.repo_get_all_commits(self._owner, self._name, **kwargs)
        assert isinstance(commits, list)

        if author:
            commits = [
                c
                for c in commits
                if author in (c.commit.author.name or c.commit.author.email)
            ]

        return [giteatools.create_commit_model(commit) for commit in commits]

    def get_workflow(self, workflow_id: str) -> Workflow:
        raise NotImplementedError

    def list_workflows(self) -> list[Workflow]:
        raise NotImplementedError

    def get_workflow_run(self, run_id: str) -> WorkflowRun:
        raise NotImplementedError

    @giteatools.handle_api_errors("Failed to download file")
    def download(
        self,
        path: str | os.PathLike[str],
        destination: str | os.PathLike[str],
        recursive: bool = False,
    ) -> None:
        """Download repository contents."""
        dest = upath.UPath(destination)
        dest.mkdir(exist_ok=True, parents=True)

        if recursive:
            contents = self._repo_api.repo_get_contents_list(
                self._owner,
                self._name,
                str(path),
                ref=self.default_branch,
            )

            for content in contents:
                if content.type == "file":
                    file_content = self._repo_api.repo_get_contents(
                        self._owner,
                        self._name,
                        content.path,
                        ref=self.default_branch,
                    )
                    file_dest = dest / content.path
                    file_dest.parent.mkdir(exist_ok=True, parents=True)
                    file_dest.write_bytes(file_content.content.encode())
        else:
            content = self._repo_api.repo_get_contents(
                self._owner,
                self._name,
                str(path),
                ref=self.default_branch,
            )
            file_dest = dest / upath.UPath(path).name
            file_dest.write_bytes(content.content.encode())

    @giteatools.handle_api_errors("Failed to search commits")
    def search_commits(
        self,
        query: str,
        branch: str | None = None,
        path: str | None = None,
        max_results: int | None = None,
    ) -> list[Commit]:
        """Search repository commits."""
        kwargs: dict[str, Any] = {}
        if branch:
            kwargs["ref_name"] = branch
        if path:
            kwargs["path"] = path
        if max_results:
            kwargs["limit"] = max_results

        commits = self._repo_api.repo_search_commits(
            self._owner,
            self._name,
            keyword=query,
            **kwargs,
        )
        return [giteatools.create_commit_model(commit) for commit in commits]

    @giteatools.handle_api_errors("Failed to list files")
    def iter_files(
        self,
        path: str = "",
        ref: str | None = None,
        pattern: str | None = None,
    ) -> Iterator[str]:
        """Iterate over repository files."""
        entries = self._repo_api.repo_get_contents_list(
            self._owner,
            self._name,
            str(path),
            ref=ref or self.default_branch,
        )
        for entry in entries:
            if entry.type == "file":
                if not pattern or fnmatch.fnmatch(entry.path, pattern):
                    yield entry.path
            elif entry.type == "dir":
                yield from self.iter_files(entry.path, ref, pattern)

    @giteatools.handle_api_errors("Failed to get contributors")
    def get_contributors(
        self,
        sort_by: Literal["commits", "name", "date"] = "commits",
        limit: int | None = None,
    ) -> list[User]:
        """Get repository contributors."""
        commits = self._repo_api.repo_get_all_commits(self._owner, self._name)
        assert isinstance(commits, list)

        # Build contributor stats from commits
        contributors: dict[str, dict[str, Any]] = {}
        for commit in commits:
            author = commit.author
            if not author:
                continue

            if author.login not in contributors:
                contributors[author.login] = {
                    "user": author,
                    "commits": 0,
                }
            contributors[author.login]["commits"] += 1

        # Convert to list and sort
        contributor_list = list(contributors.values())
        if sort_by == "name":
            contributor_list.sort(key=lambda c: c["user"].login)
        elif sort_by == "commits":
            contributor_list.sort(key=lambda c: c["commits"], reverse=True)

        # Apply limit if specified
        if limit:
            contributor_list = contributor_list[:limit]

        return [giteatools.create_user_model(c["user"]) for c in contributor_list]

    @giteatools.handle_api_errors("Failed to get latest release")
    def get_latest_release(
        self,
        include_drafts: bool = False,
        include_prereleases: bool = False,
    ) -> Release:
        """Get the latest repository release."""
        kwargs = {
            "draft": include_drafts,
            "pre_release": include_prereleases,
        }
        releases = self._repo_api.repo_list_releases(
            self._owner, self._name, per_page=1, **kwargs
        )

        if not releases:
            msg = "No matching releases found"
            raise ResourceNotFoundError(msg)

        return giteatools.create_release_model(releases[0])

    @giteatools.handle_api_errors("Failed to list releases")
    def list_releases(
        self,
        include_drafts: bool = False,
        include_prereleases: bool = False,
        limit: int | None = None,
    ) -> list[Release]:
        """List repository releases."""
        kwargs = {"per_page": limit} if limit else {}
        results = self._repo_api.repo_list_releases(self._owner, self._name, **kwargs)
        assert isinstance(results, list)
        return [
            giteatools.create_release_model(release)
            for release in results
            if ((release.draft and include_drafts) or not release.draft)
            and ((release.prerelease and include_prereleases) or not release.prerelease)
        ]

    @giteatools.handle_api_errors("Failed to get release")
    def get_release(self, tag: str) -> Release:
        """Get a specific release by tag."""
        release = self._repo_api.repo_get_release(self._owner, self._name, tag)
        return giteatools.create_release_model(release)

    @giteatools.handle_api_errors("Failed to list tags")
    def list_tags(self) -> list[Tag]:
        """List all repository tags."""
        tags = self._repo_api.repo_list_tags(self._owner, self._name)
        return [giteatools.create_tag_model(tag) for tag in tags]

    @giteatools.handle_api_errors("Failed to get tag {name}")
    def get_tag(self, name: str) -> Tag:
        """Get a specific repository tag."""
        tag = self._repo_api.repo_get_tag(self._owner, self._name, name)
        return giteatools.create_tag_model(tag)

    def get_languages(self) -> dict[str, int]:
        raise NotImplementedError

    def compare_branches(
        self,
        base: str,
        head: str,
        include_commits: bool = True,
        include_files: bool = True,
        include_stats: bool = True,
    ) -> dict[str, Any]:
        raise NotImplementedError

    @giteatools.handle_api_errors("Failed to create pull request")
    def create_pull_request(
        self,
        title: str,
        body: str,
        head_branch: str,
        base_branch: str,
        draft: bool = False,
    ) -> PullRequest:
        # Gitea doesn't support draft PRs, so we'll ignore that parameter
        if draft:
            logger.warning("Gitea does not support draft pull requests")

        pr = self._repo_api.repo_create_pull_request(
            owner=self._owner,
            repo=self._name,
            body={
                "title": title,
                "body": body,
                "head": head_branch,
                "base": base_branch,
            },
        )
        return giteatools.create_pull_request_model(pr)


if __name__ == "__main__":
    import os

    # Test Gitea API
    gitea = GiteaRepository.from_url(url="https://gitea.com/phil65/test")
    print(gitea.default_branch)



from __future__ import annotations

import os
from unittest.mock import Mock, patch

import pytest

from githarbor.providers.gitlab_provider import GitLabRepository


TOKEN = os.getenv("GITHUB_TOKEN")
OWNER = "phil65"
REPO = "test"
BRANCH = "master"


@pytest.fixture
def mock_gitlab_client():
    return Mock()


@pytest.fixture
def gitlab_repo(mock_gitlab_client) -> GitLabRepository:
    with patch("gitlab.Gitlab") as mock_gitlab:
        mock_gitlab.return_value = mock_gitlab_client
        provider = GitLabRepository("phil65", "test")
        provider._gl = mock_gitlab_client
        return provider


def supports_url():
    assert GitLabRepository.supports_url("https://gitlab.com/owner/repo")
    assert not GitLabRepository.supports_url("https://github.com/owner/repo")


def test_initialization():
    """Test basic initialization of GitLabRepository."""
    repo = GitLabRepository(OWNER, REPO)
    assert repo._owner == OWNER
    assert repo._name == REPO


def test_from_url():
    """Test creating repository from URL."""
    url = "https://gitlab.com/phil65/test"
    repo = GitLabRepository.from_url(url)
    assert repo._owner == "phil65"
    assert repo._name == "test"


def test_invalid_url():
    """Test creating repository from invalid URL."""
    import gitlab.exceptions

    with pytest.raises(gitlab.exceptions.GitlabParsingError):
        GitLabRepository.from_url("https://invalid-url.com/owner/repo")


def test_get_default_branch(gitlab_repo: GitLabRepository, mock_gitlab_client):
    """Test getting default branch."""
    mock_project = Mock()
    mock_project.default_branch = "main"
    mock_gitlab_client.projects.get.return_value = mock_project
    # branch = gitlab_repo.get_default_branch()
    # assert branch == "main"


# def test_get_branches(gitlab_repo, mock_gitlab_client):
#     """Test getting all branches"""
#     mock_project = Mock()
#     mock_project.branches.list.return_value = [
#         Mock(name="main"),
#         Mock(name="develop")
#     ]
#     mock_gitlab_client.projects.get.return_value = mock_project

#     branches = gitlab_repo.get_branches()
#     assert len(branches) == 2
#     assert "main" in branches
#     assert "develop" in branches


# def test_get_file_content(gitlab_repo, mock_gitlab_client):
#     """Test getting file content"""
#     mock_project = Mock()
#     mock_file = Mock()
#     mock_file.decode.return_value = "file content"
#     mock_project.files.get.return_value = mock_file
#     mock_gitlab_client.projects.get.return_value = mock_project

#     content = gitlab_repo.get_file_content("path/to/file.txt", "main")
#     assert content == "file content"


def test_supports_url_variations():
    """Test URL support with different variations."""
    assert GitLabRepository.supports_url("https://gitlab.com/owner/repo")
    assert GitLabRepository.supports_url("http://gitlab.com/owner/repo")
    assert GitLabRepository.supports_url("gitlab.com/owner/repo")
    assert not GitLabRepository.supports_url("https://github.com/owner/repo")
    assert not GitLabRepository.supports_url("https://bitbucket.org/owner/repo")


# def test_get_repo_url(gitlab_repo):
#     """Test getting repository URL"""
#     expected_url = f"https://gitlab.com/{OWNER}/{REPO}"
#     assert gitlab_repo.get_repo_url() == expected_url

# @pytest.mark.parametrize("invalid_input", [
#     ("", "repo"),
#     ("owner", ""),
#     (None, "repo"),
#     ("owner", None)
# ])
# def test_invalid_initialization(invalid_input):
#     """Test initialization with invalid inputs"""
#     owner, repo = invalid_input
#     with pytest.raises(ValueError):
#         GitLabRepository(owner, repo)



"""Module containing GitHubKit owner implementation."""

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any, ClassVar

from githubkit import GitHub
from githubkit.exception import GitHubException

from githarbor.core.base import BaseOwner, BaseRepository
from githarbor.exceptions import AuthenticationError, ResourceNotFoundError
from githarbor.providers.githubkit_provider import utils as githubkittools
from githarbor.repositories import create_repository


if TYPE_CHECKING:
    from githarbor.core.models import User

logger = logging.getLogger(__name__)


class GitHubKitOwner(BaseOwner):
    """Owner implementation using GitHubKit."""

    url_patterns: ClassVar[list[str]] = ["github.com"]
    is_async = True

    def __init__(self, username: str, token: str | None = None) -> None:
        """Initialize GitHub API client.

        Args:
            username: Username of the owner
            token: GitHub access token
        """
        try:
            t = token or os.getenv("GITHUB_TOKEN")
            if not t:
                msg = "GitHub token is required"
                raise ValueError(msg)
            self._token = t
            self._gh = GitHub(t)
            self._name = username
            self._user = self._gh.rest.users.get_authenticated().parsed_data

        except GitHubException as e:
            msg = f"GitHub authentication failed: {e!s}"
            raise AuthenticationError(msg) from e

    @property
    def name(self) -> str:
        """The name of the owner."""
        return self._name

    @classmethod
    def from_url(cls, url: str, **kwargs: Any) -> GitHubKitOwner:
        """Create instance from URL.

        Args:
            url: URL like "https://github.com/octocat"
            **kwargs: Additional arguments including token
        """
        # Remove trailing slashes and handle both user & org URLs
        url = url.rstrip("/")
        username = url.split("/")[-1]
        return cls(username=username, token=kwargs.get("token"))

    async def list_repos_owned_by_user(
        self, headers: dict[str, str] | None = None
    ) -> list[dict[str, Any]]:
        """List repositories owned by user.

        Args:
            headers: Optional additional headers to pass

        Returns:
            List of repository data
        """
        import aiohttp

        base_headers = {
            "Accept": "application/vnd.github.v3+json",
            "Authorization": f"Bearer {self._token}",
        }
        if headers:
            base_headers.update(headers)

        url = f"https://api.github.com/users/{self._name}/repos"

        async with aiohttp.ClientSession() as session:  # noqa: SIM117
            async with session.get(url, headers=base_headers) as response:
                if response.status == 200:  # noqa: PLR2004
                    return await response.json()
                msg = f"Failed to list repositories: {response.status}"
                raise ResourceNotFoundError(msg)

    async def list_repositories_async(self) -> list[BaseRepository]:
        """List all repositories owned by this user."""
        try:
            repos = await self.list_repos_owned_by_user()
            return [
                create_repository(repo["html_url"])
                for repo in repos
                if repo.get("html_url")
            ]
        except Exception as e:
            msg = f"Failed to list repositories: {e!s}"
            raise ResourceNotFoundError(msg) from e

    async def create_repository_async(
        self,
        name: str,
        description: str = "",
        private: bool = False,
    ) -> BaseRepository:
        """Create a new repository.

        Args:
            name: Repository name
            description: Repository description
            private: Whether to create a private repository

        Returns:
            Newly created repository
        """
        try:
            resp = await self._gh.rest.repos.async_create_for_authenticated_user(
                name=name,
                description=description,
                private=private,
            )
            return create_repository(resp.parsed_data.html_url)
        except GitHubException as e:
            msg = f"Failed to create repository: {e!s}"
            raise ResourceNotFoundError(msg) from e

    async def get_user_async(self) -> User:
        """Get user information."""
        try:
            resp = await self._gh.rest.users.async_get_by_username(self._name)
            return githubkittools.create_user_model(resp.parsed_data)
        except GitHubException as e:
            msg = f"Failed to get user info: {e!s}"
            raise ResourceNotFoundError(msg) from e

    async def delete_repository_async(self, name: str) -> None:
        """Delete a repository.

        Args:
            name: Name of repository to delete
        """
        try:
            await self._gh.rest.repos.async_delete(self._name, name)
        except GitHubException as e:
            msg = f"Failed to delete repository: {e!s}"
            raise ResourceNotFoundError(msg) from e



from githarbor.core.base import BaseRepository
from githarbor.core.models import (
    Branch,
    Comment,
    Commit,
    Issue,
    Label,
    PullRequest,
    Release,
    User,
    Workflow,
    WorkflowRun,
)
from githarbor.exceptions import (
    AuthenticationError,
    GitHarborError,
    OperationNotAllowedError,
    ProviderNotConfiguredError,
    RateLimitError,
    RepositoryNotFoundError,
    ResourceNotFoundError,
)
from githarbor.repositories import create_repository

__version__ = "0.6.2"

__all__ = [
    "AuthenticationError",
    # Base
    "BaseRepository",
    # Models
    "Branch",
    "Comment",
    "Commit",
    # Exceptions
    "GitHarborError",
    "Issue",
    "Label",
    "OperationNotAllowedError",
    "ProviderNotConfiguredError",
    "PullRequest",
    "RateLimitError",
    "Release",
    "RepositoryNotFoundError",
    "ResourceNotFoundError",
    "User",
    "Workflow",
    "WorkflowRun",
    # Factory
    "create_repository",
]



from __future__ import annotations

import importlib.util
from typing import TYPE_CHECKING, Any

from githarbor.exceptions import OwnerNotFoundError
from githarbor.registry import RepoRegistry


if TYPE_CHECKING:
    from githarbor.core.user_proxy import Owner


if importlib.util.find_spec("github"):
    from githarbor.providers.github_provider.owner import GitHubOwner

    RepoRegistry.register_owner("github")(GitHubOwner)

if importlib.util.find_spec("aiogithubapi"):
    from githarbor.providers.aiogithubapi_provider.owner import AioGitHubOwner

    RepoRegistry.register_owner("aiogithubapi")(AioGitHubOwner)

if importlib.util.find_spec("githubkit"):
    from githarbor.providers.githubkit_provider.owner import GitHubKitOwner

    RepoRegistry.register_owner("githubkit")(GitHubKitOwner)


def create_owner(url: str, **kwargs: Any) -> Owner:
    """Create a proxy-wrapped owner instance from a URL.

    Args:
        url: The owner URL (e.g. 'https://github.com/phil65')
        **kwargs: Provider-specific configuration (tokens, etc.)

    Returns:
        Owner: Proxy-wrapped owner instance

    Raises:
        OwnerNotFoundError: If the URL isn't supported or no owner found
    """
    try:
        return RepoRegistry.owner_from_url(url, **kwargs)
    except Exception as e:
        msg = f"Failed to create owner from {url}: {e!s}"
        raise OwnerNotFoundError(msg) from e


if __name__ == "__main__":
    owner = create_owner("https://github.com/phil65")
    print(owner)



"""Githarbor providers package."""

from __future__ import annotations

import importlib.util

if importlib.util.find_spec("github"):
    from githarbor.providers.github_provider import GitHubRepository

if importlib.util.find_spec("gitlab"):
    from githarbor.providers.gitlab_provider import GitLabRepository

if importlib.util.find_spec("giteapy"):
    from githarbor.providers.gitea_provider import GiteaRepository

if importlib.util.find_spec("azure"):
    from githarbor.providers.azure_provider import AzureRepository

if importlib.util.find_spec("aiogithubapi"):
    from githarbor.providers.aiogithubapi_provider import AioGitHubRepository

if importlib.util.find_spec("githubkit"):
    from githarbor.providers.githubkit_provider import GitHubKitRepository

if importlib.util.find_spec("gitpython"):
    from githarbor.providers.local_provider import LocalRepository


# if importlib.util.find_spec("atlassian"):
#     from githarbor.providers.bitbucketrepository import BitbucketRepository

__all__ = [
    "AioGitHubRepository",
    "AzureRepository",
    "GitHubKitRepository",
    "GitHubRepository",
    "GitLabRepository",
    "GiteaRepository",
    "LocalRepository",
    # "BitbucketRepository",
]



import os

import pytest

from githarbor import create_repository


BITBUCKET_TOKEN = os.getenv("BITBUCKET_TOKEN")
GITHUB_TOKEN = os.getenv("GITHUB_TOKEN")
GITHUB_URL = "https://github.com/phil65/mknodes"
GITLAB_TOKEN = os.getenv("GITLAB_TOKEN")


@pytest.mark.integration
@pytest.mark.skipif(not os.getenv("GITHUB_TOKEN"), reason="GITHUB_TOKEN not set")
def test_github_integration():
    repo = create_repository(GITHUB_URL, token=GITHUB_TOKEN)
    assert repo.name == "mknodes"  # Updated to match actual repo name


@pytest.mark.integration
def test_gitlab_integration():
    repo = create_repository("https://gitlab.com/phil65/test", token=GITLAB_TOKEN)
    assert repo.name == "test"


# @pytest.mark.integration
# def test_bitbucket_integration():
#     repo = create_repository(
#         "https://bitbucket.org/phil__65/testrepo/", token=BITBUCKET_TOKEN
#     )
#     assert repo.name == "testrepo"



from __future__ import annotations

import fnmatch
import logging
import os
from typing import TYPE_CHECKING, Any, ClassVar, Literal
from urllib.parse import urlparse

from githarbor.core.base import BaseRepository, IssueState, PullRequestState
from githarbor.exceptions import AuthenticationError, ResourceNotFoundError
from githarbor.providers.github_provider import utils as githubtools


if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

    from githarbor.core.models import (
        Branch,
        Commit,
        Issue,
        PullRequest,
        Release,
        Tag,
        User,
        Workflow,
        WorkflowRun,
    )


logger = logging.getLogger(__name__)
HTML_ERROR_CODE = 404
TOKEN = os.getenv("GITHUB_TOKEN")


class GitHubRepository(BaseRepository):
    """GitHub repository implementation."""

    url_patterns: ClassVar[list[str]] = ["github.com"]
    raw_prefix: ClassVar[str] = (
        "https://raw.githubusercontent.com/{owner}/{name}/{branch}/{path}"
    )

    def __init__(self, owner: str, name: str, token: str | None = None):
        """Initialize GitHub repository."""
        from github import Auth, Github, NamedUser
        from github.GithubException import GithubException

        try:
            t = token or TOKEN
            if t is None:
                logger.info("No GitHub token provided. Stricter rate limit.")
            auth = Auth.Token(t) if t else None
            self._gh = Github(auth=auth)
            self._repo = self._gh.get_repo(f"{owner}/{name}")
            self._owner = owner
            self._name = name
            self.user: NamedUser.NamedUser = self._gh.get_user(owner)  # type: ignore
        except GithubException as e:
            msg = f"GitHub authentication failed: {e!s}"
            raise AuthenticationError(msg) from e

    @classmethod
    def from_url(cls, url: str, **kwargs: Any) -> GitHubRepository:
        """Create from URL like 'https://github.com/owner/repo'."""
        parsed = urlparse(url)
        parts = parsed.path.strip("/").split("/")
        if len(parts) < 2:  # noqa: PLR2004
            msg = f"Invalid GitHub URL: {url}"
            raise ValueError(msg)

        return cls(parts[0], parts[1].removesuffix(".git"), token=kwargs.get("token"))

    @property
    def default_branch(self) -> str:
        return self._repo.default_branch

    @property
    def edit_base_uri(self):
        return f"edit/{self.default_branch}/"

    @githubtools.handle_github_errors("Failed to get branch {name}")
    def get_branch(self, name: str) -> Branch:
        branch = self._repo.get_branch(name)
        model = githubtools.create_branch_model(branch)
        model.default = branch.name == self.default_branch
        return model

    @githubtools.handle_github_errors("Failed to get pull request {number}")
    def get_pull_request(self, number: int) -> PullRequest:
        pr = self._repo.get_pull(number)
        return githubtools.create_pull_request_model(pr)

    @githubtools.handle_github_errors("Failed to list pull requests")
    def list_pull_requests(self, state: PullRequestState = "open") -> list[PullRequest]:
        prs = self._repo.get_pulls(state=state)
        return [githubtools.create_pull_request_model(pr) for pr in prs]

    @githubtools.handle_github_errors("Failed to get issue {issue_id}")
    def get_issue(self, issue_id: int) -> Issue:
        issue = self._repo.get_issue(issue_id)
        return githubtools.create_issue_model(issue)

    @githubtools.handle_github_errors("Failed to list issues")
    def list_issues(self, state: IssueState = "open") -> list[Issue]:
        issues = self._repo.get_issues(state=state)
        return [githubtools.create_issue_model(issue) for issue in issues]

    @githubtools.handle_github_errors("Failed to get commit {sha}")
    def get_commit(self, sha: str) -> Commit:
        commit = self._repo.get_commit(sha)
        return githubtools.create_commit_model(commit)

    @githubtools.handle_github_errors("Failed to list commits")
    def list_commits(
        self,
        branch: str | None = None,
        since: datetime | None = None,
        until: datetime | None = None,
        author: str | None = None,
        path: str | None = None,
        max_results: int | None = None,
    ) -> list[Commit]:
        kwargs = {
            "since": since,
            "until": until,
            "author": author,
            "path": path,
            "sha": branch,
        }
        # Filter out None values
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        commits = self._repo.get_commits(**kwargs)  # type: ignore
        results = commits[:max_results] if max_results else commits
        return [githubtools.create_commit_model(c) for c in results]

    @githubtools.handle_github_errors("Failed to get workflow {workflow_id}")
    def get_workflow(self, workflow_id: str) -> Workflow:
        workflow = self._repo.get_workflow(workflow_id)
        return githubtools.create_workflow_model(workflow)

    @githubtools.handle_github_errors("Failed to list workflows")
    def list_workflows(self) -> list[Workflow]:
        workflows = self._repo.get_workflows()
        return [githubtools.create_workflow_model(w) for w in workflows]

    @githubtools.handle_github_errors("Failed to get workflow run {run_id}")
    def get_workflow_run(self, run_id: str) -> WorkflowRun:
        run = self._repo.get_workflow_run(int(run_id))
        return githubtools.create_workflow_run_model(run)

    @githubtools.handle_github_errors("Failed to download file {path}")
    def download(
        self,
        path: str | os.PathLike[str],
        destination: str | os.PathLike[str],
        recursive: bool = False,
    ):
        user_name = self._gh.get_user().login if TOKEN else None
        return githubtools.download_from_github(
            org=self._owner,
            repo=self._name,
            path=path,
            destination=destination,
            username=user_name,
            token=TOKEN,
            recursive=recursive,
        )

    @githubtools.handle_github_errors("Failed to search commits")
    def search_commits(
        self,
        query: str,
        branch: str | None = None,
        path: str | None = None,
        max_results: int | None = None,
    ) -> list[Commit]:
        # Build the search query
        search_query = f"{query} repo:{self._owner}/{self._name}"
        # Add branch qualifier if specified
        if branch:
            search_query += f" ref:{branch}"
        # Add path qualifier if specified
        if path:
            search_query += f" path:{path}"
        kwargs = {"query": search_query}
        # kwargs = {"query": f"{self._owner}/{self._name}+{query}"}
        # if branch:
        #     kwargs["ref"] = branch
        # if path:
        #     kwargs["path"] = path
        results = self._gh.search_commits(**kwargs)
        commits = list(results[:max_results] if max_results else results)
        return [self.get_commit(c.sha) for c in commits]

    @githubtools.handle_github_errors("Failed to list files for {path}")
    def iter_files(
        self,
        path: str = "",
        ref: str | None = None,
        pattern: str | None = None,
    ) -> Iterator[str]:
        contents = self._repo.get_contents(path, ref=ref or self.default_branch)
        assert isinstance(contents, list)
        kwargs = {"ref": ref} if ref else {}
        while contents:
            content = contents.pop(0)
            if content.type == "dir":
                c = self._repo.get_contents(content.path, **kwargs)
                assert isinstance(c, list)
                contents.extend(c)
            elif not pattern or fnmatch.fnmatch(content.path, pattern):
                yield content.path

    @githubtools.handle_github_errors("Failed to get repository owner info")
    def get_repo_user(self) -> User:
        """Get user (repository owner) information."""
        return githubtools.create_user_model(self.user)

    @githubtools.handle_github_errors("Failed to get contributors")
    def get_contributors(
        self,
        sort_by: Literal["commits", "name", "date"] = "commits",
        limit: int | None = None,
    ) -> list[User]:
        contributors = list(self._repo.get_contributors())
        if sort_by == "name":
            contributors = sorted(contributors, key=lambda c: c.login)
        elif sort_by == "date":
            contributors = sorted(contributors, key=lambda c: c.created_at)
        contributors = contributors[:limit] if limit else contributors
        return [u for c in contributors if (u := githubtools.create_user_model(c))]

    @githubtools.handle_github_errors("Failed to get languages")
    def get_languages(self) -> dict[str, int]:
        return self._repo.get_languages()

    @githubtools.handle_github_errors("Failed to compare branches")
    def compare_branches(
        self,
        base: str,
        head: str,
        include_commits: bool = True,
        include_files: bool = True,
        include_stats: bool = True,
    ) -> dict[str, Any]:
        comparison = self._repo.compare(base, head)
        result: dict[str, Any] = {
            "ahead_by": comparison.ahead_by,
            "behind_by": comparison.behind_by,
        }

        if include_commits:
            result["commits"] = [self.get_commit(c.sha) for c in comparison.commits]
        if include_files:
            result["files"] = [f.filename for f in comparison.files]
        if include_stats:
            result["stats"] = {
                "additions": comparison.total_commits,
                "deletions": comparison.total_commits,
                "changes": len(comparison.files),
            }
        return result

    @githubtools.handle_github_errors("Failed to get latest release")
    def get_latest_release(
        self,
        include_drafts: bool = False,
        include_prereleases: bool = False,
    ) -> Release:  # Changed from dict[str, Any] to Release
        releases = self._repo.get_releases()
        # Filter releases based on parameters
        filtered = [
            release
            for release in releases
            if (include_drafts or not release.draft)
            and (include_prereleases or not release.prerelease)
        ]
        if not filtered:
            msg = "No matching releases found"
            raise ResourceNotFoundError(msg)
        latest = filtered[0]  # Releases are returned in chronological order
        return githubtools.create_release_model(latest)

    @githubtools.handle_github_errors("Failed to list releases")
    def list_releases(
        self,
        include_drafts: bool = False,
        include_prereleases: bool = False,
        limit: int | None = None,
    ) -> list[Release]:
        filtered_releases = (
            release
            for release in self._repo.get_releases()
            if (include_drafts or not release.draft)
            and (include_prereleases or not release.prerelease)
        )
        return [
            githubtools.create_release_model(release)
            for release in (
                list(filtered_releases)[:limit] if limit else filtered_releases
            )
        ]

    @githubtools.handle_github_errors("Failed to get release {tag}")
    def get_release(self, tag: str) -> Release:
        release = self._repo.get_release(tag)
        return githubtools.create_release_model(release)

    @githubtools.handle_github_errors("Failed to get tag {name}")
    def get_tag(self, name: str) -> Tag:
        """Get a specific tag by name."""
        from github.GithubException import GithubException

        try:
            tag = self._repo.get_git_ref(f"tags/{name}")
            tag_obj = self._repo.get_git_tag(tag.object.sha)
            return githubtools.create_tag_model(tag_obj)
        except GithubException as e:
            if e.status == HTML_ERROR_CODE:  # Might be lightweight tag
                commit = self._repo.get_commit(name)
                return githubtools.create_tag_model(commit)
            raise

    @githubtools.handle_github_errors("Failed to list tags")
    def list_tags(self) -> list[Tag]:
        """List all repository tags."""
        return [githubtools.create_tag_model(tag) for tag in self._repo.get_tags()]

    @githubtools.handle_github_errors("Failed to create pull request")
    def create_pull_request(
        self,
        title: str,
        body: str,
        head_branch: str,
        base_branch: str,
        draft: bool = False,
    ) -> PullRequest:
        pr = self._repo.create_pull(
            title=title,
            body=body,
            head=head_branch,
            base=base_branch,
            draft=draft,
        )
        return githubtools.create_pull_request_model(pr)

    @githubtools.handle_github_errors("Failed to create pull request from diff")
    def create_pull_request_from_diff(
        self,
        title: str,
        body: str,
        base_branch: str,
        diff: str,
        head_branch: str | None = None,
        draft: bool = False,
    ) -> PullRequest:
        """Create a pull request from a diff string."""
        from datetime import datetime
        import uuid

        from github import InputGitTreeElement
        from github.GithubException import GithubException

        from githarbor.core.filechanges import parse_diff
        from githarbor.exceptions import GitHarborError

        # Generate unique branch name if not provided
        if head_branch is None:
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            unique_id = str(uuid.uuid4())[:8]
            head_branch = f"patch/{timestamp}_{unique_id}"

        # Get the base branch's last commit
        base_ref = self._repo.get_git_ref(f"refs/heads/{base_branch}")
        base_commit = self._repo.get_git_commit(base_ref.object.sha)

        # Create a new branch
        try:
            head_ref = self._repo.get_git_ref(f"refs/heads/{head_branch}")
            msg = f"Branch {head_branch} already exists"
            raise GitHarborError(msg)
        except GithubException:
            ref = f"refs/heads/{head_branch}"
            head_ref = self._repo.create_git_ref(ref=ref, sha=base_ref.object.sha)

        # Parse the diff and apply changes
        changes = parse_diff(diff)

        # Create blobs and trees for the changes
        new_tree: list[InputGitTreeElement] = []
        for change in changes:
            if change.mode == "delete":
                # For deletions, we add a null SHA
                elem = InputGitTreeElement(
                    path=change.path,
                    mode="100644",
                    type="blob",
                    sha=None,
                )
                new_tree.append(elem)
                continue

            if change.content is not None:
                # Create blob for the file content
                blob = self._repo.create_git_blob(
                    content=change.content, encoding="utf-8"
                )

                if change.old_path:
                    # For renamed files, we need to remove the old path
                    elem = InputGitTreeElement(
                        path=change.old_path,
                        mode="100644",
                        type="blob",
                        sha=None,
                    )
                    new_tree.append(elem)

                elem = InputGitTreeElement(
                    path=change.path,
                    mode="100644",
                    type="blob",
                    sha=blob.sha,
                )
                new_tree.append(elem)

        # Create a new tree
        base_tree = self._repo.get_git_tree(base_commit.tree.sha)
        tree = self._repo.create_git_tree(new_tree, base_tree)

        # Create a commit
        msg = f"Changes for {title}"
        commit = self._repo.create_git_commit(msg, tree=tree, parents=[base_commit])

        # Update the reference
        head_ref.edit(commit.sha, force=True)

        # Create the pull request
        pr = self._repo.create_pull(
            title=title,
            body=body,
            base=base_branch,
            head=head_branch,
            draft=draft,
        )
        return githubtools.create_pull_request_model(pr)

    @githubtools.handle_github_errors("Failed to create branch")
    def create_branch(
        self,
        name: str,
        base_commit: str,
    ) -> Branch:
        """Create a new branch at the specified commit."""
        # Create reference with full ref name
        _ref = self._repo.create_git_ref(
            ref=f"refs/heads/{name}",
            sha=base_commit,
        )
        # Get the branch to return proper Branch model
        branch = self._repo.get_branch(name)
        return githubtools.create_branch_model(branch)


if __name__ == "__main__":
    repo = GitHubRepository.from_url("https://github.com/phil65/mknodes")
    commits = repo.search_commits("implement")
    print(commits)
    # print(repo.list_workflows())
    branch = repo.get_branch("main")
    print(branch)



from __future__ import annotations

import os
from unittest.mock import Mock, patch

import pytest

from githarbor.exceptions import ResourceNotFoundError
from githarbor.providers.github_provider import GitHubRepository


TOKEN = os.getenv("GITHUB_TOKEN")
OWNER = "phil65"
REPO = "mknodes"
BRANCH = "main"


@pytest.fixture
def mock_github():
    with patch("github.Github") as mock:
        yield mock


@pytest.fixture
def github_repo(mock_github):
    # Setup mock repository
    mock_repo = Mock()
    mock_repo.name = REPO
    mock_repo.default_branch = BRANCH

    mock_github.return_value.get_repo.return_value = mock_repo

    return GitHubRepository(OWNER, REPO, token=TOKEN)


def test_github_repository_init(mock_github):
    repo = GitHubRepository(OWNER, REPO, token=TOKEN)
    assert repo.name == "mknodes"
    mock_github.assert_called_once()


def test_github_repository_from_url(mock_github):
    repo = GitHubRepository.from_url(f"https://github.com/{OWNER}/{REPO}", token=TOKEN)
    assert repo.name == "mknodes"


def test_get_branch(github_repo):
    # Setup mock branch
    mock_branch = Mock()
    mock_branch.name = "main"
    mock_branch.commit.sha = "abc123"
    mock_branch.protected = False

    github_repo._repo.get_branch.return_value = mock_branch

    branch = github_repo.get_branch("main")
    assert branch.name == "main"
    assert branch.sha == "abc123"
    assert branch.protected is False


def test_branch_not_found(github_repo):
    from github.GithubException import GithubException

    github_repo._repo.get_branch.side_effect = GithubException(404, "Not Found")

    with pytest.raises(ResourceNotFoundError):
        github_repo.get_branch("nonexistent")
